from dataclasses import dataclass

import plotly.express as px

from psdm_analysis.models.result.grid.enhanced_node import EnhancedNodesResult
from psdm_analysis.models.result.grid.node import NodesResult


@dataclass(frozen=True)
class NodesResultPlotter:
    results: NodesResult

    def boxplot_v_mag(self, uuid: str):
        node_res = self.results.entities.get(uuid)
        return px.box(node_res.v_mag)

    def boxplots_v_mag(self):
        return px.box(self.results.v_mag)

    def boxplot_v_ang(self, uuid: str):
        node_res = self.results.entities.get(uuid)
        return px.box(node_res.v_ang)

    def boxplots_v_ang(self):
        return px.box(self.results.v_ang)

    def lineplots_v_mag(self):
        return px.line(self.results.v_mag)

    def lineplots_v_ang(self):
        return px.line(self.results.v_ang)


@dataclass(frozen=True)
class EnhancedNodesResultPlotter(NodesResultPlotter):
    results: EnhancedNodesResult

    def lineplots_p(self):
        return px.line(self.results.p)

    def boxplots_p(self):
        return px.box(self.results.p)



from enum import Enum
from typing import TypeVar


class EntitiesEnum(Enum):
    def has_type(self):
        if self in {
            RawGridElementsEnum.LINE,
            RawGridElementsEnum.TRANSFORMER_2_W,
            RawGridElementsEnum.TRANSFROMER_3_W,
            SystemParticipantsEnum.ELECTRIC_VEHICLE,
            SystemParticipantsEnum.BIOMASS_PLANT,
            SystemParticipantsEnum.WIND_ENERGY_CONVERTER,
            SystemParticipantsEnum.STORAGE,
            SystemParticipantsEnum.HEAT_PUMP,
        }:
            return True
        else:
            return False

    def get_csv_input_file_name(self):
        return self.value + "_input.csv"

    def get_csv_result_file_name(self):
        return self.value + "_res.csv"

    def get_type_file_name(self):
        assert self.has_type() is True
        return self.value + "_type_input.csv"

    def get_plot_name(self):
        return self.value.replace("_", " ").title()

    def get_result_type(self):
        # locally to avoid circular imports
        from psdm_analysis.models.result.grid.connector import ConnectorResult
        from psdm_analysis.models.result.grid.node import NodeResult
        from psdm_analysis.models.result.grid.transformer import Transformer2WResult
        from psdm_analysis.models.result.power import PQResult, PQWithSocResult

        if isinstance(self, SystemParticipantsEnum):
            if self.has_soc():
                return PQWithSocResult
            else:
                return PQResult
        elif isinstance(self, RawGridElementsEnum):
            match self:
                case RawGridElementsEnum.NODE:
                    return NodeResult
                case RawGridElementsEnum.TRANSFORMER_2_W:
                    return Transformer2WResult
                case RawGridElementsEnum.LINE:
                    return ConnectorResult
                case _:
                    raise NotImplementedError(
                        f"Result type {self} not implemented yet!"
                    )
        else:
            raise ValueError(f"Entity type {self} not supported!")


EntityEnumType = TypeVar("EntityEnumType", bound=EntitiesEnum)


class SystemParticipantsEnum(EntitiesEnum):
    BIOMASS_PLANT = "bm"
    COMBINED_HEAT_AND_POWER = "chp"
    ELECTRIC_VEHICLE = "ev"
    EV_CHARGING_STATION = "evcs"
    FIXED_FEED_IN = "fixed_feed_in"
    LOAD = "load"
    PHOTOVOLTAIC_POWER_PLANT = "pv"
    WIND_ENERGY_CONVERTER = "wec"
    STORAGE = "storage"
    ENERGY_MANAGEMENT = "em"
    HEAT_PUMP = "hp"
    FLEX_OPTIONS = "flex_options"
    PRIMARY_DATA = "primary_data"
    PARTICIPANTS_SUM = "participants_sum"

    @staticmethod
    def values():
        return [participant for participant in SystemParticipantsEnum]

    def has_soc(self):
        return self in {
            SystemParticipantsEnum.ELECTRIC_VEHICLE,
            SystemParticipantsEnum.STORAGE,
        }


class RawGridElementsEnum(EntitiesEnum):
    NODE = "node"
    LINE = "line"
    TRANSFORMER_2_W = "transformer_2_w"
    TRANSFROMER_3_W = "transformer_3_w"
    SWITCH = "switch"
    MEASUREMENT_UNIT = "measurement_unit"



from abc import ABC
from dataclasses import dataclass

import pandas as pd

from psdm_analysis.models.entity import Entities
from psdm_analysis.models.input.node import Nodes


@dataclass(frozen=True)
class Connector(Entities, ABC):
    @staticmethod
    def attributes() -> list[str]:
        return Entities.attributes() + ["node_a", "node_b", "parallel_devices"]

    @property
    def node_a(self):
        return self.data["node_a"]

    @property
    def node_b(self):
        return self.data["node_b"]

    @property
    def parallel_devices(self):
        return self.data["parallel_devices"]

    def nodes(self):
        return pd.concat([self.node_a, self.node_b])

    def filter_for_node(self, uuid: str):
        data = self.data[(self.node_a == uuid) | (self.node_b == uuid)]
        return type(self)(data)

    def insert_node_id_columns(self, nodes: Nodes) -> None:
        index_to_id = nodes.ids.to_dict()

        if "node_a_id" not in self.data.columns:
            self.data.insert(self.data.columns.get_loc("node_a") + 1, "node_a_id", None)
        self.data["node_a_id"] = self.node_a.map(index_to_id)

        if "node_b_id" not in self.data.columns:
            self.data.insert(self.data.columns.get_loc("node_b") + 1, "node_b_id", None)
        self.data["node_b_id"] = self.node_b.map(index_to_id)



import re
from dataclasses import dataclass
from enum import Enum

import pandas as pd


class CurrentType(Enum):
    AC = "ac"
    DC = "dc"


@dataclass(frozen=True)
class ChargingPointType:
    type_id: str
    # in kW
    power: float
    current_type: CurrentType
    synonymous_ids: set[str]


HOUSEHOLD_SOCKET = ChargingPointType(
    "HouseholdSocket", 2.3, CurrentType.AC, {"household", "hhs", "schuko-simple"}
)

BLUE_HOUSEHOLD_SOCKET = ChargingPointType(
    "BlueHouseholdSocket",
    3.6,
    CurrentType.AC,
    {"bluehousehold", "bhs", "schuko-camping"},
)

CEE16A_SOCKET = ChargingPointType("Cee16ASocket", 11, CurrentType.AC, {"cee16"})

CEE32A_SOCKET = ChargingPointType("Cee32ASocket", 22, CurrentType.AC, {"cee32"})

CEE63A_SOCKET = ChargingPointType("Cee63ASocket", 43, CurrentType.AC, {"cee63"})

CHARGING_STATION_TYPE_1 = ChargingPointType(
    "ChargingStationType1", 7.2, CurrentType.AC, {"cst1", "stationtype1", "cstype1"}
)

CHARGING_STATION_TYPE_2 = ChargingPointType(
    "ChargingStationType2", 43, CurrentType.AC, {"cst2", "stationtype2", "cstype2"}
)

CHARGING_STATION_CCS_COMBO_TYPE_1 = ChargingPointType(
    "ChargingStationCcsComboType1", 11, CurrentType.DC, {"csccs1", "csccscombo1"}
)

CHARGING_STATION_CCS_COMBO_TYPE_2 = ChargingPointType(
    "ChargingStationCcsComboType2", 50, CurrentType.DC, {"csccs2", "csccscombo2"}
)

TESLA_SUPER_CHARGER_V1 = ChargingPointType(
    "TeslaSuperChargerV1",
    135,
    CurrentType.DC,
    {"tesla1", "teslav1", "supercharger1", "supercharger"},
)

TESLA_SUPER_CHARGER_V2 = ChargingPointType(
    "TeslaSuperChargerV2",
    150,
    CurrentType.DC,
    {"tesla2", "teslav2", "supercharger2", "supercharger2"},
)

TESLA_SUPER_CHARGER_V3 = ChargingPointType(
    "TeslaSuperChargerV3",
    250,
    CurrentType.DC,
    {"tesla3", "teslav3", "supercharger3", "supercharger3"},
)

common_charging_point_types = [
    HOUSEHOLD_SOCKET,
    BLUE_HOUSEHOLD_SOCKET,
    CEE16A_SOCKET,
    CEE32A_SOCKET,
    CEE63A_SOCKET,
    CHARGING_STATION_TYPE_1,
    CHARGING_STATION_TYPE_2,
    CHARGING_STATION_CCS_COMBO_TYPE_1,
    CHARGING_STATION_CCS_COMBO_TYPE_2,
    TESLA_SUPER_CHARGER_V1,
    TESLA_SUPER_CHARGER_V2,
    TESLA_SUPER_CHARGER_V3,
]


def get_common_charging_point_type(id: str):
    for charging_point_type in common_charging_point_types:
        if (
            id == charging_point_type.type_id
            or id in charging_point_type.synonymous_ids
        ):
            return charging_point_type
    return


def parse_evcs_type(type_str: str) -> ChargingPointType:
    common_type = get_common_charging_point_type(type_str)
    if common_type:
        return common_type

    regex = r"\(.*\)$"
    match = re.search(regex, type_str)
    if match:
        power, current_type = match.group().strip("()").split("|")
        current_type = CurrentType[current_type]
        return ChargingPointType(type_str, float(power), current_type, set())
    else:
        raise ValueError(f"Can not determine power of {type_str}!")


def parse_evcs_type_info(type_str: str):
    evcs_type = parse_evcs_type(type_str)
    return pd.Series(evcs_type.__dict__)



import logging
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Optional, Type, TypeVar, Union

from pandas import DataFrame
from pandas.core.groupby import DataFrameGroupBy

from psdm_analysis.io.utils import (
    check_filter,
    csv_to_grpd_df,
    get_file_path,
    to_date_time,
)
from psdm_analysis.models.entity import Entities, ResultEntities
from psdm_analysis.models.input.enums import EntitiesEnum, EntityEnumType

ResultDictType = TypeVar("ResultDictType", bound="ResultDict")


@dataclass(frozen=True)
class ResultDict(ABC):
    entity_type: EntitiesEnum
    entities: Dict[str, ResultEntities]

    def __len__(self):
        return len(self.entities)

    def __contains__(self, uuid):
        return uuid in self.entities

    def __getitem__(self, get):
        match get:
            case str():
                return self.entities[get]
            case slice():
                start, stop, step = get.start, get.stop, get.step
                if step is not None:
                    logging.warning("Step is not supported for slicing. Ignoring it.")
                if not (isinstance(start, datetime) and isinstance(stop, datetime)):
                    raise ValueError("Only datetime slicing is supported")
                entities = {
                    key: e.filter_for_time_interval(start, stop)
                    for key, e in self.entities.items()
                }
                return type(self)(self.entity_type, entities)
            case _:
                raise ValueError(
                    "Only get by uuid or datetime slice for filtering is supported."
                )

    @classmethod
    def from_csv(
        cls: Type[ResultDictType],
        entity_type: EntityEnumType,
        simulation_data_path: str,
        delimiter: str,
        simulation_end: Optional[datetime] = None,
        input_entities: Optional[Entities] = None,
        filter_start: Optional[datetime] = None,
        filter_end: Optional[datetime] = None,
    ) -> ResultDictType:
        check_filter(filter_start, filter_end)
        grpd_df = ResultDict.get_grpd_df(
            entity_type,
            simulation_data_path,
            delimiter,
        )
        if not grpd_df:
            logging.debug("There are no " + str(cls))
            return cls.create_empty(entity_type)
        if simulation_end is None:
            simulation_end = to_date_time(grpd_df["time"].max().max())
        entities = dict(
            grpd_df.apply(
                lambda grp: ResultDict.build_for_entity(
                    entity_type,
                    grp.name,
                    grp.drop(columns=["input_model"]),
                    simulation_end,
                    input_entities=input_entities,
                )
            )
        )
        res = cls(
            entity_type,
            entities,
        )
        return (
            res
            if not filter_start
            else res.filter_for_time_interval(filter_start, filter_end)
        )

    @staticmethod
    def build_for_entity(
        entity_type: EntityEnumType,
        input_model: str,
        data: DataFrame,
        simulation_end: datetime,
        input_entities=Optional[Entities],
    ) -> ResultDictType:
        name = None
        if input_entities is not None:
            if input_model not in input_entities.ids.index:
                logging.debug(
                    f"Input model {input_model} of type {entity_type} not found in input entities. It seems like the wrong input_entities have been passed. Not assigning a name to the result."
                )
            else:
                name = input_entities.ids.loc[input_model]

        return entity_type.get_result_type().build(
            entity_type, input_model, data, simulation_end, name=name
        )

    @staticmethod
    def get_grpd_df(
        entity_type: EntityEnumType,
        simulation_data_path: str,
        delimiter: str,
    ) -> Optional[DataFrameGroupBy]:
        file_name = entity_type.get_csv_result_file_name()
        path = get_file_path(simulation_data_path, file_name)

        if not path.exists():
            logging.info(
                "No results built for {} since {} does not exist".format(
                    file_name, str(path)
                )
            )
            return None

        return csv_to_grpd_df(file_name, simulation_data_path, delimiter)

    @staticmethod
    def safe_get_path(entity_type: EntityEnumType, data_path: str) -> Optional[str]:
        file_name = entity_type.get_csv_result_file_name()
        path = get_file_path(data_path, file_name)
        if path.exists():
            return path
        else:
            logging.info(
                "No results built for {} since {} does not exist".format(
                    file_name, str(path)
                )
            )
            return None

    @classmethod
    def create_empty(cls, sp_type):
        return cls(sp_type, dict())

    def uuids(self):
        return list(self.entities.keys())

    def results(self):
        return list(self.entities.values())

    # noinspection PyArgumentList
    def subset(self, uuids):
        matched_participants = {
            uuid: self.entities[uuid] for uuid in self.entities.keys() & uuids
        }

        return type(self)(self.entity_type, matched_participants)

    def subset_split(self, uuids: list[str]):
        """
        Returns a results result containing the given uuids and a results result containing the remaining uuids.
        :param uuids: the uuids with which to split the result
        :return:
        """

        rmd_uuids = self.entities.keys() - uuids
        return self.subset(uuids), self.subset(rmd_uuids)

    def filter_by_date_time(self, time: Union[datetime, list[datetime]]):
        """
        Filters the result by the given datetime or list of datetimes.
        :param time: the time or list of times to filter by
        :return: a new result containing only the given time or times
        """
        return type(self)(
            self.entity_type,
            {uuid: result[time] for uuid, result in self.entities.items()},
        )

    # noinspection PyArgumentList
    def filter_for_time_interval(self, start: datetime, end: datetime):
        return type(self)(
            self.entity_type,
            {
                uuid: result.filter_for_time_interval(start, end)
                for uuid, result in self.entities.items()
            },
        )

    def uuid_to_id_map(self) -> dict[str, Optional[str]]:
        return {uuid: result.name for uuid, result in self.entities.items()}



from dataclasses import dataclass

from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class EnergyManagementSystems(SystemParticipants):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.ENERGY_MANAGEMENT

    @property
    def connected_assets(self):
        return self.data["connected_assets"]

    @property
    def control_strategy(self):
        return self.data["control_strategy"]

    def uuid_to_connected_assets(self):
        return self.connected_assets.to_dict()



import concurrent.futures
import logging
import os
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from typing import Union

import pandas as pd
from pandas import Series

from psdm_analysis.io import utils
from psdm_analysis.io.utils import to_date_time
from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.result.power import PQResult


@dataclass
class PrimaryData:
    name: str
    # ts_uuid -> ts
    time_series: dict[str, PQResult]
    # participant_uuid -> ts_uuid
    participant_mapping: dict[str, str]

    @property
    def p(self):
        if not self.time_series.values():
            return None
        return (
            pd.DataFrame({p_uuid: res.p for p_uuid, res in self.time_series.items()})
            .fillna(method="ffill")
            .sort_index()
        )

    def sum(self) -> PQResult:
        return PQResult.sum(list(self.time_series.values()))

    def p_sum(self) -> Series:
        if not self.time_series:
            return Series(dtype=float)
        return self.p.fillna(method="ffill").sum(axis=1).rename("p_sum")

    @property
    def q(self):
        return (
            pd.DataFrame({p_uuid: res.q for p_uuid, res in self.time_series.items()})
            .fillna(method="ffill")
            .sort_index()
        )

    def q_sum(self):
        if not self.time_series:
            return Series(dtype=float)
        return self.q.fillna(method="ffill").sum(axis=1).rename("q_sum")

    def get_for_participant(self, participant: str):
        time_series_id = self.participant_mapping[participant]
        return self.time_series[time_series_id]

    def get_for_participants(self, participants):
        time_series = []
        for participant in participants:
            if participant in self.participant_mapping:
                time_series.append(self.get_for_participant(participant))
        return time_series

    def filter_by_date_time(self, time: Union[datetime, list[datetime]]):
        """
        Filters the result by the given datetime or list of datetimes.
        :param time: the time or list of times to filter by
        :return: a new result containing only the given time or times
        """
        return PrimaryData(
            self.name,
            {uuid: result[time] for uuid, result in self.time_series.items()},
            self.participant_mapping,
        )

    def filter_for_time_interval(self, start: datetime, end: datetime):
        filtered_time_series = {
            uuid: time_series.filter_for_time_interval(start, end)
            for uuid, time_series in self.time_series.items()
        }
        return PrimaryData(self.name, filtered_time_series, self.participant_mapping)

    @classmethod
    def from_csv(cls, path: str, delimiter: str):
        # get all files that start with "its_"
        path = utils.get_absolute_path(path)
        ts_files = [file for file in os.listdir(path) if file.startswith("its_p")]

        time_series_dict = {}

        if ts_files:
            ts_mapping = utils.read_csv(str(path), "time_series_mapping.csv", delimiter)

            participant_mapping = (
                ts_mapping[["participant", "time_series"]]
                .set_index("participant")
                .to_dict()["time_series"]
            )

            pa_read_time_series = partial(PrimaryData.read_time_series, path, delimiter)

            with concurrent.futures.ProcessPoolExecutor() as executor:
                time_series = executor.map(pa_read_time_series, ts_files)
                for ts in time_series:
                    time_series_dict[ts.name] = ts

            return PrimaryData("Primary Data", time_series_dict, participant_mapping)

        else:
            logging.debug(f"No primary data in path {path}")
            return PrimaryData("Primary Data", dict(), dict())

    @staticmethod
    def read_time_series(dir_path: str, delimiter: str, ts_file: str):
        name = ts_file[-40:][:36]
        data = utils.read_csv(str(dir_path), ts_file, delimiter)
        data["time"] = data["time"].apply(lambda date_string: to_date_time(date_string))
        data = data.set_index("time")
        if "q" not in data.columns:
            data["q"] = 0
        return PQResult(SystemParticipantsEnum.PRIMARY_DATA, name, name, data)



import dataclasses
import os
import re
from dataclasses import dataclass
from functools import reduce
from typing import List, Tuple, Union

import pandas as pd
from pandas import Series

from psdm_analysis.io.utils import get_absolute_path
from psdm_analysis.models.entity import ResultEntities
from psdm_analysis.models.input.enums import EntitiesEnum, SystemParticipantsEnum
from psdm_analysis.processing.dataframe import divide_positive_negative
from psdm_analysis.processing.series import (
    add_series,
    duration_weighted_sum,
    hourly_mean_resample,
    load_and_generation,
)


@dataclass(frozen=True)
class PQResult(ResultEntities):
    def __add__(self, other: "PQResult"):
        p_sum = add_series(self.p, other.p, "p")
        q_sum = add_series(self.q, other.q, "q")
        summed_data = p_sum.to_frame().join(q_sum)
        return PQResult(
            self.type
            if self.type == other.type
            else SystemParticipantsEnum.PARTICIPANTS_SUM,
            "PQResult - Sum",
            "PQResult - Sum",
            summed_data,
        )

    def __sub__(self, other: "PQResult"):
        return self + (other * -1)

    def __mul__(self, other: Union[float, int]):
        if isinstance(other, float) or isinstance(other, int):
            updated_data = self.data * other
            return PQResult(self.type, self.input_model, self.name, updated_data)
        else:
            raise ValueError(
                f"Multiplication with type {type(other)} not or not yet supported"
            )

    # commutative multiplication (a*b=b*a)
    __rmul__ = __mul__

    def __eq__(self, other):
        if not isinstance(other, PQResult):
            return False
        return (self.data == other.data).all()

    @classmethod
    def from_csv(cls, file_path: str, sp_type: EntitiesEnum, name: str = None):
        file_path = get_absolute_path(file_path)
        data = pd.read_csv(file_path)
        data["time"] = pd.to_datetime(data["time"])
        data = data.set_index("time", drop=True)
        file_name = os.path.basename(file_path)
        is_default_file_name = re.match(
            f"{sp_type.get_csv_result_file_name()}_.*", file_name
        )
        input_model = file_name.split("_")[-1] if is_default_file_name else file_name
        return cls(sp_type, name if name else input_model, input_model, data)

    def to_csv(self, path: str, file_name: str = None, delimiter: str = ","):
        file_name = file_name if file_name else self.get_default_output_name()
        self.data.to_csv(os.path.join(path, file_name), sep=delimiter)

    def get_default_output_name(self):
        return self.input_model + "_" + self.type.get_csv_result_file_name()

    @property
    def p(self):
        return self.data["p"]

    @property
    def q(self):
        return self.data["q"]

    @staticmethod
    def attributes() -> List[str]:
        return ["p", "q"]

    @classmethod
    # todo: find a way for parallel calculation
    def sum(cls, results: List["PQResult"]) -> "PQResult":
        if len(results) == 0:
            return PQResult.create_empty(
                SystemParticipantsEnum.PARTICIPANTS_SUM, "", ""
            )
        if len(results) == 1:
            return results[0]
        return reduce(lambda a, b: a + b, results)

    def energy(self) -> float:
        return duration_weighted_sum(self.p)

    def load_and_generation(self) -> Tuple[float, float]:
        return load_and_generation(self.p)

    def divide_load_generation(self):
        load, generation = divide_positive_negative(self.data)
        return dataclasses.replace(self, data=load), dataclasses.replace(
            self, data=generation
        )

    def daily_usage(self, device_power_mw):
        return (
            self.p()
            .groupby(self.p.index.date)
            .apply(lambda series: duration_weighted_sum(series.abs()))
            / device_power_mw
        )

    def annual_duration_series(self, drop_index=True):
        return (
            hourly_mean_resample(self.p)
            .sort_values(ascending=False)
            .reset_index(drop=drop_index)
        )

    def hourly_resample(self):
        updated_data = self.data.apply(lambda col: hourly_mean_resample(col), axis=0)
        return PQResult(self.type, self.input_model, self.name, updated_data)

    def complex_power(self):
        return self.p + 1j * self.q


@dataclass(frozen=True)
class PQWithSocResult(PQResult):
    def add_with_soc(
        self, this_capacity: float, other: "PQWithSocResult", other_capacity: float
    ):
        p_sum = add_series(self.p, other.p, "p")
        q_sum = add_series(self.q, other.q, "q")
        total_capacity = this_capacity + other_capacity
        soc = add_series(
            self.soc() * this_capacity / total_capacity,
            other.soc() * other_capacity / total_capacity,
            "soc",
        )
        summed_data = p_sum.to_frame().join([q_sum, soc])
        return PQWithSocResult(
            self.type,
            "PQResult - Sum",
            "PQResult - Sum",
            summed_data,
        )

    @staticmethod
    # todo: find a way for parallel calculation
    def sum_with_soc(
        results: list[Tuple[float, "PQWithSocResult"]]
    ) -> "PQWithSocResult":
        if len(results) == 0:
            return PQWithSocResult.create_empty(
                SystemParticipantsEnum.PARTICIPANTS_SUM, "", ""
            )
        if len(results) == 1:
            return results[0][1]
        this_capacity, agg = results[0]
        for other_capacity, result in results[1::]:
            agg = agg.add_with_soc(this_capacity, result, other_capacity)
        return agg

    def soc(self) -> Series:
        return self.data["soc"]

    @staticmethod
    def attributes() -> List[str]:
        return ["p", "q", "capacity"]



import logging
from dataclasses import dataclass

import pandas as pd

from psdm_analysis.models.input.container.mixins import ContainerMixin
from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.input.participant.bm import BiomassPlants
from psdm_analysis.models.input.participant.em import EnergyManagementSystems
from psdm_analysis.models.input.participant.evcs import EvChargingStations
from psdm_analysis.models.input.participant.evs import ElectricVehicles
from psdm_analysis.models.input.participant.fixed_feed_in import FixedFeedIns
from psdm_analysis.models.input.participant.hp import HeatPumps
from psdm_analysis.models.input.participant.load import Loads
from psdm_analysis.models.input.participant.pv import PhotovoltaicPowerPlants
from psdm_analysis.models.input.participant.storage import Storages
from psdm_analysis.models.input.participant.wec import WindEnergyConverters


@dataclass(frozen=True)
class SystemParticipantsContainer(ContainerMixin):
    ems: EnergyManagementSystems
    loads: Loads
    fixed_feed_ins: FixedFeedIns
    pvs: PhotovoltaicPowerPlants
    biomass_plants: BiomassPlants
    wecs: WindEnergyConverters
    storages: Storages
    evs: ElectricVehicles
    evcs: EvChargingStations
    hps: HeatPumps

    @staticmethod
    def from_csv(path: str, delimiter: str) -> "SystemParticipantsContainer":
        loads = Loads.from_csv(path, delimiter)
        fixed_feed_ins = FixedFeedIns.from_csv(path, delimiter)
        pvs = PhotovoltaicPowerPlants.from_csv(path, delimiter)
        biomass_plants = BiomassPlants.from_csv(path, delimiter)
        wecs = WindEnergyConverters.from_csv(path, delimiter)
        storages = Storages.from_csv(path, delimiter)
        ems = EnergyManagementSystems.from_csv(path, delimiter)
        evs = ElectricVehicles.from_csv(path, delimiter)
        evcs = EvChargingStations.from_csv(path, delimiter)
        hps = HeatPumps.from_csv(path, delimiter)
        return SystemParticipantsContainer(
            ems,
            loads,
            fixed_feed_ins,
            pvs,
            biomass_plants,
            wecs,
            storages,
            evs,
            evcs,
            hps,
        )

    def to_list(self, include_empty=False):
        participants = [
            self.ems,
            self.loads,
            self.fixed_feed_ins,
            self.pvs,
            self.biomass_plants,
            self.wecs,
            self.storages,
            self.evs,
            self.evcs,
            self.hps,
        ]
        return (
            participants
            if include_empty
            else [p for p in participants if not p.data.empty]
        )

    def filter_by_node(self, node_uuid: str):
        loads = self.loads.filter_for_node(node_uuid)
        fixed_feed_ins = self.fixed_feed_ins.filter_for_node(node_uuid)
        pvs = self.pvs.filter_for_node(node_uuid)
        biomass_plants = self.biomass_plants.filter_for_node(node_uuid)
        wecs = self.wecs.filter_for_node(node_uuid)
        storages = self.storages.filter_for_node(node_uuid)
        ems = self.ems.filter_for_node(node_uuid)
        evs = self.evs.filter_for_node(node_uuid)
        evcs = self.evcs.filter_for_node(node_uuid)
        hps = self.hps.filter_for_node(node_uuid)
        return SystemParticipantsContainer(
            ems,
            loads,
            fixed_feed_ins,
            pvs,
            biomass_plants,
            wecs,
            storages,
            evs,
            evcs,
            hps,
        )

    def get_participants(self, sp_type: SystemParticipantsEnum):
        if sp_type == SystemParticipantsEnum.ENERGY_MANAGEMENT:
            return self.ems
        elif sp_type == SystemParticipantsEnum.LOAD:
            return self.loads
        elif sp_type == SystemParticipantsEnum.BIOMASS_PLANT:
            return self.biomass_plants
        elif sp_type == SystemParticipantsEnum.FIXED_FEED_IN:
            return self.fixed_feed_ins
        elif sp_type == SystemParticipantsEnum.PHOTOVOLTAIC_POWER_PLANT:
            return self.pvs
        elif sp_type == SystemParticipantsEnum.WIND_ENERGY_CONVERTER:
            return self.wecs
        elif sp_type == SystemParticipantsEnum.STORAGE:
            return self.storages
        elif sp_type == SystemParticipantsEnum.EV_CHARGING_STATION:
            return self.evcs
        elif sp_type == SystemParticipantsEnum.ELECTRIC_VEHICLE:
            return self.evs
        elif sp_type == SystemParticipantsEnum.HEAT_PUMP:
            return self.hps
        else:
            logging.error(f"No return value for system participant of type: {sp_type}")
            return None

    def find_participant(self, uuid: str):
        if uuid in self.loads:
            return self.loads.get(uuid)
        elif uuid in self.fixed_feed_ins:
            return self.fixed_feed_ins.get(uuid)
        elif uuid in self.pvs:
            return self.pvs.get(uuid)
        elif uuid in self.wecs:
            return self.wecs.get(uuid)
        elif uuid in self.biomass_plants:
            return self.biomass_plants.get(uuid)
        elif uuid in self.ems:
            return self.ems.get(uuid)
        elif uuid in self.storages:
            return self.storages.get(uuid)
        elif uuid in self.evs:
            return self.evs.get(uuid)
        elif uuid in self.evcs:
            return self.evcs.get(uuid)
        elif uuid in self.hps:
            return self.hps.get(uuid)
        else:
            raise ValueError(
                f"Particpant with uuid {uuid} could not be found among participants."
            )

    def uuids(self):
        return pd.concat(
            [
                participants.uuids.to_series()
                for participants in self.to_list(include_empty=True)
            ]
        )

    def subset(self, uuids):
        return SystemParticipantsContainer(
            self.ems.subset(uuids),
            self.loads.subset(uuids),
            self.fixed_feed_ins.subset(uuids),
            self.pvs.subset(uuids),
            self.wecs.subset(uuids),
            self.biomass_plants.subset(uuids),
            self.storages.subset(uuids),
            self.evs.subset(uuids),
            self.evcs.subset(uuids),
            self.hps.subset(uuids),
        )



import os
from datetime import datetime
from pathlib import Path
from typing import Optional

import pandas as pd
from pandas import DataFrame
from pandas.core.groupby import DataFrameGroupBy

ROOT_DIR = os.path.abspath(__file__ + "/../../../")


def get_absolute_path(path: str):
    if path.startswith(ROOT_DIR):
        return path
    else:
        return Path(ROOT_DIR).joinpath(path)


def get_file_path(path: str, file_name: str):
    if path.startswith(ROOT_DIR):
        return Path(path).joinpath(file_name)
    else:
        return Path(ROOT_DIR).joinpath(path).joinpath(file_name)


def read_csv(path: str, file_name: str, delimiter: str) -> DataFrame:
    full_path = get_file_path(path, file_name)
    if not full_path.exists():
        raise IOError("File with path: " + str(full_path) + " does not exist")
    return pd.read_csv(full_path, delimiter=delimiter, quotechar='"')


def to_date_time(zoned_date_time: str):
    """
    Converts zoned date time string with format: "yyyy-MM-dd'T'HH:mm:ss[.S[S][S]]'Z'"
    e.g. '2022-02-01T00:15Z[UTC]' to python datetime
    :param zoned_date_time:
    :return:
    """
    if not zoned_date_time or not isinstance(zoned_date_time, str):
        raise ValueError(f"Unexpected date time string: {zoned_date_time}")
    year = int(zoned_date_time[0:4])
    month = int(zoned_date_time[5:7])
    day = int(zoned_date_time[8:10])
    hour = int(zoned_date_time[11:13])
    minute = int(zoned_date_time[14:16])
    return datetime(year=year, month=month, day=day, hour=hour, minute=minute)


def csv_to_grpd_df(
    file_name: str, simulation_data_path: str, delimiter: str
) -> DataFrameGroupBy:
    """
    Reads in a PSDM csv results file cleans it up and groups it by input_archive model.

    :param file_name: name of the file to read
    :param simulation_data_path: base directory of the result data
    :param delimiter: the csv delimiter
    :return: DataFrameGroupBy object of the file
    """
    data = read_csv(simulation_data_path, file_name, delimiter)

    if "uuid" in data.columns:
        data = data.drop(columns=["uuid"])
    return data.groupby(by="input_model")


def check_filter(filter_start: Optional[datetime], filter_end: Optional[datetime]):
    if (filter_start or filter_end) and not (filter_start and filter_end):
        raise ValueError(
            "Both start and end of the filter must be provided if one is provided."
        )
    if (filter_start and filter_end) and (filter_start > filter_end):
        raise ValueError("Filter start must be before end.")


def df_to_csv(df: DataFrame, path: str, file_name: str, delimiter: str):
    full_path = get_file_path(path, file_name)
    df = df.replace(True, "true")
    df = df.replace(False, "false")
    df = df.sort_index()
    df.to_csv(full_path, index=True, index_label="uuid", sep=delimiter)



from __future__ import annotations

import json
import logging
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Tuple, TypeVar, Union

import pandas as pd
from pandas import DataFrame, Series

from psdm_analysis.io import utils
from psdm_analysis.io.utils import df_to_csv, read_csv, to_date_time
from psdm_analysis.models.input.enums import (
    EntitiesEnum,
    RawGridElementsEnum,
    SystemParticipantsEnum,
)
from psdm_analysis.models.input.participant.charging import parse_evcs_type_info
from psdm_analysis.processing.dataframe import filter_data_for_time_interval

if TYPE_CHECKING:
    from psdm_analysis.models.input.node import Nodes

EntityType = TypeVar("EntityType", bound="Entities")


@dataclass(frozen=True)
class Entities(ABC):
    data: DataFrame

    def __len__(self):
        return len(self.data)

    def __contains__(self, uuid: str):
        return uuid in self.data.index

    @classmethod
    def from_csv(cls, path: str, delimiter: str):
        return cls._from_csv(path, delimiter, cls.get_enum())

    @staticmethod
    @abstractmethod
    def get_enum() -> EntitiesEnum:
        pass

    def to_csv(self, path: str, delimiter: str = ","):
        df_to_csv(self.data, path, self.get_enum().get_csv_input_file_name(), delimiter)

    @classmethod
    def _from_csv(cls, path: str, delimiter: str, entity: EntitiesEnum):
        file_path = utils.get_file_path(path, entity.get_csv_input_file_name())
        if os.path.exists(file_path):
            return cls(Entities._data_from_csv(entity, path, delimiter))
        else:
            logging.debug(
                "There is no file named: "
                + str(file_path)
                + ". No "
                + entity.value
                + " entities are loaded."
            )
            return cls.create_empty()

    @classmethod
    def _data_from_csv(
        cls, entity: EntitiesEnum, path: str, delimiter: str
    ) -> DataFrame:
        data = read_csv(path, entity.get_csv_input_file_name(), delimiter)
        if entity.has_type():
            type_data = read_csv(path, entity.get_type_file_name(), delimiter)
            data = data.merge(
                type_data, left_on="type", right_on="uuid", suffixes=("", "_type")
            ).rename(columns={"id_type": "type_id", "uuid_type": "type_uuid"})
        # special data transformations
        match entity:
            # for raw grid elements
            # ---------------------
            case RawGridElementsEnum.NODE:
                data["longitude"] = data["geo_position"].apply(
                    lambda geo_json: json.loads(geo_json)["coordinates"][0]
                )
                data["latitude"] = data["geo_position"].apply(
                    lambda geo_json: json.loads(geo_json)["coordinates"][1]
                )

            # for system participants
            # -----------------------
            case SystemParticipantsEnum.ENERGY_MANAGEMENT:
                data["connected_assets"] = data["connected_assets"].apply(
                    lambda x: x.split(" ")
                )
            case SystemParticipantsEnum.EV_CHARGING_STATION:
                type_data = data["type"].apply(
                    lambda type_str: parse_evcs_type_info(type_str)
                )
                data = pd.concat([data, type_data])

        try:
            return data.set_index("uuid")
        except KeyError as e:
            raise KeyError(
                "Column 'uuid' not found. This might be due to wrong csv delimiter!", e
            )

    @property
    def uuids(self):
        return self.data.index

    @property
    def ids(self):
        return self.data["id"]

    @property
    def operates_from(self):
        return self.data["operates_from"]

    @property
    def operates_until(self):
        return self.data["operates_until"]

    @property
    def operator(self):
        return self.data["operator"]

    def __repr__(self):
        return self.data

    def get(self, uuid: str) -> Series:
        return self.data.loc[uuid]

    @staticmethod
    def attributes() -> List[str]:
        """
        Method that should hold all attributes field (transformed to snake_case and case-sensitive)
        of the corresponding PSDM entity
        :return:
        """
        return ["uuid", "id", "operates_from", "operates_until", "operator"]

    @classmethod
    def create_empty(cls):
        data = pd.DataFrame(columns=cls.attributes())
        return cls(data)

    def subset(self, uuids: Union[list[str], str]):
        if isinstance(uuids, str):
            uuids = [uuids]
        return type(self)(self.data.loc[uuids])

    def subset_split(self, uuids: list[str]):
        rmd = set(self.uuids) - set(uuids)
        return self.subset(uuids), self.subset(list(rmd))

    @abstractmethod
    def nodes(self):
        pass

    def filter_for_node(self, uuid: str):
        data = self.data[self.nodes() == str(uuid)]
        return type(self)(data)

    def find_nodes(self, nodes: Nodes) -> Nodes:
        return nodes.subset(self.nodes())


ResultType = TypeVar("ResultType", bound="ResultEntities")


@dataclass(frozen=True)
class ResultEntities(ABC):
    # todo: type is a reserved keyword -> rename
    type: EntitiesEnum
    input_model: str
    name: Optional[str]
    data: DataFrame

    def __repr__(self):
        return self.data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, where: Union[slice, datetime, list[datetime]]):
        if isinstance(where, slice):
            start, stop, step = where.start, where.stop, where.step
            if step is not None:
                logging.warning("Step is not supported for slicing. Ignoring it.")
            if not (isinstance(start, datetime) and isinstance(stop, datetime)):
                raise ValueError("Only datetime slicing is supported")
            return self.filter_for_time_interval(start, stop)
        elif isinstance(where, datetime):
            filtered, dt = self._get_data_by_datetime(where)
            data = pd.DataFrame(filtered).T
            return self.build(self.type, self.input_model, data, dt, self.name)
        elif isinstance(where, list):
            filtered = [self._get_data_by_datetime(time)[0] for time in where]
            data = pd.DataFrame(pd.concat(filtered, axis=1)).T
            max_dt = sorted(data.index)[-1]
            return self.build(self.type, self.input_model, data, max_dt, name=self.name)

    def _get_data_by_datetime(self, dt: datetime) -> Tuple[Series, datetime]:
        if dt > self.data.index[-1]:
            logging.warning(
                "Trying to access data after last time step. Returning last time step."
            )
            return self.data.iloc[-1], self.data.index[-1]
        if dt < self.data.index[0]:
            logging.warning(
                "Trying to access data before first time step. Returning first time step."
            )
            return self.data.iloc[0], self.data.index[0]
        else:
            return self.data.asof(dt), dt

    @staticmethod
    @abstractmethod
    def attributes() -> List[str]:
        pass

    @classmethod
    def empty_data(cls, index=None):
        return (
            pd.DataFrame(columns=cls.attributes(), index=index)
            if index
            else pd.DataFrame(columns=cls.attributes())
        )

    @classmethod
    def create_empty(
        cls, entity_type: EntitiesEnum, name: Optional[str], input_model: str
    ):
        data = cls.empty_data()
        return cls(entity_type, input_model, name, data)

    @classmethod
    def build(
        cls,
        entity_type: EntitiesEnum,
        input_model: str,
        data: DataFrame,
        end: datetime,
        name: Optional[str] = None,
    ) -> "ResultEntities":
        if data.empty:
            return cls.create_empty(entity_type, name, input_model)

        if "time" in data.columns:
            data["time"] = data["time"].apply(
                lambda date_string: to_date_time(date_string)
            )
            data = data.set_index("time", drop=True)

        last_state = data.iloc[len(data) - 1]
        if last_state.name != end:
            last_state.name = end
            data = pd.concat([data, DataFrame(last_state).transpose()])
        # todo: deal with duplicate indexes -> take later one
        data = data[~data.index.duplicated(keep="last")]
        data.sort_index(inplace=True)
        return cls(entity_type, input_model, name, data)

    def filter_for_time_interval(self, start: datetime, end: datetime):
        filtered_data = filter_data_for_time_interval(self.data, start, end)
        return self.build(self.type, self.input_model, filtered_data, end, self.name)

    @classmethod
    # todo: find a way for parallel calculation
    def sum(cls, results: list[ResultType]) -> ResultType:
        if len(results) == 0:
            return cls.create_empty(SystemParticipantsEnum.PARTICIPANTS_SUM, "", "")
        if len(results) == 1:
            return results[0]
        agg = results[0]
        for result in results[1::]:
            agg += result
        return agg

    def find_input_entity(self, input_model: EntityType):
        if self.type != input_model.get_enum():
            logging.warning("Input model type does not match result type!")
        return input_model.subset([self.input_model])



import math
from datetime import datetime
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.axes import Axes

from psdm_analysis.models.grid_with_results import GridWithResults
from psdm_analysis.models.result.grid.enhanced_node import EnhancedNodesResult
from psdm_analysis.models.result.grid.node import NodeResult, NodesResult
from psdm_analysis.plots.common.line_plot import ax_plot_time_series
from psdm_analysis.plots.common.utils import (
    FIGSIZE,
    LABEL_PAD,
    set_style,
    set_subplot_title,
    set_title,
    set_xlabels_rotated,
    set_ylabel,
)

set_style()


def plot_all_v_mag_branch_violin(
    gwr: GridWithResults,
):
    """
    Plots violin plots for all nodes across all branches.

    Args:
        gwr: GridWithResults object

    Returns:
        fig, axes
    """
    branches = gwr.grid.raw_grid.get_branches()
    nodes_res = gwr.results.nodes
    width, height = FIGSIZE
    height = height * len(branches)
    fig, axes = plt.subplots(nrows=len(branches), figsize=(width, height))
    for i, branch in enumerate(branches):
        ax_plot_v_mag_branch_violin(axes[i], nodes_res, branch)
        set_subplot_title(axes[i], f"Voltages along Branch {i+1}")
    plt.tight_layout()

    return fig, axes


def plot_v_mags_violin(
    nodes_res: Union[NodesResult, EnhancedNodesResult],
    nodes: Optional[list[str]],
):
    """
    Plots violin plots for all nodes across the given branches.

    Args:
        nodes_res: NodesResult or EnhancedNodesResult object
        nodes: Optional list of node uuids that should be plotted. Order is preserved.

    Returns:
        fig, ax
    """
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_v_mag_branch_violin(ax, nodes_res, nodes)
    set_title(ax, "Voltage Magnitudes along Branch")
    return fig, ax


def ax_plot_v_mag_branch_violin(
    ax: Axes,
    nodes_res: Union[NodesResult, EnhancedNodesResult],
    nodes: Optional[list[str]],  # branches can be found by GridContainer.get_branches()
):
    """
    Plots violin plots for given nodes. If no nodes are passed all nodes are plotted.

    Args:
        ax: Axes object
        nodes_res: NodesResult or EnhancedNodesResult object
        nodes: Optional list of node uuids that should be plotted. Order is preserved.
    """

    if nodes:
        # get v_mag in listed sequence
        v_mag = nodes_res.subset(nodes).v_mag.reindex(columns=nodes)
    else:
        v_mag = nodes_res.v_mag

    ax.violinplot(v_mag, showmedians=True)

    # set labels
    uuid_to_id = nodes_res.uuid_to_id_map()
    x_labels = v_mag.columns.map(lambda uuid: uuid_to_id[uuid])
    set_xlabels_rotated(ax, x_labels, ha="right")
    set_ylabel(ax, "Voltage Magnitude in pu")
    _ = ax.set_xticklabels(x_labels, rotation=45, ha="right")


def plot_v_mag_branch(
    nodes_res: Union[NodesResult, EnhancedNodesResult],
    branch: list[str],  # branches can be found by GridContainer.get_branches()
    time: datetime,
    in_kw: bool = False,
):
    """
    Plots voltage magnitudes of nodes along the given branch.

    Args:
        nodes_res: NodesResult or EnhancedNodesResult object
        branch: list of node uuids that form the branch
        time: datetime for which the voltage magnitudes should be plotted
        in_kw: if True, power is plotted in kW instead of MW

    Returns:
        fig, ax
    """
    fig, ax = plt.subplots(figsize=(10, 5))
    ax_plot_v_mag_branch(ax, nodes_res, branch, time, fig=fig, in_kw=in_kw)
    set_title(ax, "Voltage Magnitudes along Branch")
    return fig, ax


def ax_plot_v_mag_branch(
    ax: Axes,
    nodes_res: Union[NodeResult, EnhancedNodesResult],
    branch: list[str],
    time: datetime,
    fig: Optional[plt.Figure] = None,  # used to plot colorbar
    in_kw: bool = False,
):
    """
    Plots voltage magnitudes of nodes along the given branch.

    Args:
        ax: Axes object
        nodes_res: NodesResult or EnhancedNodesResult object
        branch: list of node uuids that form the branch
        time: datetime for which the voltage magnitudes should be plotted
        fig: Optional Figure object which is used to plot the colorbar
        in_kw: if True, power is plotted in kW instead of MW
    """

    with_power = isinstance(nodes_res, EnhancedNodesResult)

    v_mags = []
    x_ticks = []
    p_s = []

    for node in branch:
        node_res = nodes_res[node]
        res = node_res[time]
        v_mag = res.v_mag.iloc[0]

        # happens for auxiliary node at open switch
        if math.isnan(v_mag):
            continue
        v_mags.append(v_mag)

        if with_power:
            p_s.append(res.p.iloc[0])
        x_ticks.append(node_res.name)

    ax.plot(x_ticks, v_mags)
    ax.set_xticks(np.arange(len(x_ticks)))  # xtick locations
    ax.set_xticklabels(x_ticks, rotation=45, ha="right")  # set xtick labels
    set_ylabel(ax, "Voltage Magnitude in p.u.")

    if with_power:
        # plot aggregated node power indicator
        cmap = sns.color_palette("coolwarm", as_cmap=True)
        if in_kw:
            p_s = [p * 1000 for p in p_s]
        sc = ax.scatter(x_ticks, v_mags, c=p_s, cmap=cmap, edgecolor="none", zorder=10)
        if fig:
            cbar = fig.colorbar(sc, ax=ax)
            unit = "kW" if in_kw else "MW"
            cbar.set_label(f"Nodal Power in {unit}", labelpad=LABEL_PAD)


def plot_v_mag(
    res: NodeResult,
    resolution: str,
    title: Optional[str] = None,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if title is None:
        title = f"Voltage Magnitude at Node: {res.name}"
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_node_v_mag(
        ax,
        res,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    set_title(ax, title)
    return fig, ax


def plot_v_ang(
    res: NodeResult,
    resolution: str,
    title: Optional[str] = None,
    hourly_mean: bool = False,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if title is None:
        title = f"Voltage Angle at Node: {res.name}"
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_node_v_ang(
        ax,
        res,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    set_title(ax, title)
    return fig, ax


def ax_plot_node_v_mag(
    ax: Axes,
    res: NodeResult,
    resolution: str,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.v_mag) == 0:
        raise ValueError("Voltage magnitude time series is empty. No data to plot")

    ax = ax_plot_time_series(
        ax,
        res.v_mag,
        res.type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Voltage Magnitude in p.u.")


def ax_plot_node_v_ang(
    ax: Axes,
    res: NodeResult,
    resolution: str,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.v_ang) == 0:
        raise ValueError("Voltage angle time series is empty. No data to plot")

    ax = ax_plot_time_series(
        ax,
        res.v_ang,
        res.type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Voltage Angle in deg")



from dataclasses import dataclass

from psdm_analysis.models.input.container.mixins import SpTypeMixin
from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class HeatPumps(SystemParticipants, SpTypeMixin):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.HEAT_PUMP

    @property
    def thermal_bus(self):
        return self.data["thermal_bus"]

    @property
    def p_thermal(self):
        return self.data["p_thermal"]

    @staticmethod
    def entity_attributes() -> list[str]:
        return ["thermal_bus"]

    @staticmethod
    def type_attributes() -> list[str]:
        return SpTypeMixin.type_attributes() + ["p_thermal"]



import concurrent.futures
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from typing import Optional, Union

import pandas as pd
from pandas import DataFrame, Series

from psdm_analysis.io.utils import check_filter
from psdm_analysis.models.input.container.grid_container import GridContainer
from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.result.participant.flex_options import FlexOptionsResult
from psdm_analysis.models.result.participant.participant import (
    ParticipantsResult,
    ParticipantsWithSocResult,
)
from psdm_analysis.models.result.power import PQResult


@dataclass(frozen=True)
class ParticipantsResultContainer:
    ems: ParticipantsResult
    loads: ParticipantsResult
    fixed_feed_ins: ParticipantsResult
    pvs: ParticipantsResult
    wecs: ParticipantsResult
    storages: ParticipantsWithSocResult
    evcs: ParticipantsResult
    evs: ParticipantsWithSocResult
    hps: ParticipantsResult
    flex: FlexOptionsResult

    def __len__(self):
        participants = self.to_list(include_empty=False)
        return sum([len(participant) for participant in participants])

    def __getitem__(self, slice_val: slice):
        if not isinstance(slice_val, slice):
            raise ValueError("Only slicing is supported!")
        start, stop, _ = slice_val.start, slice_val.stop, slice_val.step
        if not (isinstance(start, datetime) and isinstance(stop, datetime)):
            raise ValueError("Only datetime slicing is supported")
        return self.filter_for_time_interval(start, stop)

    @classmethod
    def from_csv(
        cls,
        simulation_data_path: str,
        delimiter: str,
        simulation_end: datetime,
        grid_container: Optional[GridContainer] = None,
        filter_start: Optional[datetime] = None,
        filter_end: Optional[datetime] = None,
    ):
        check_filter(filter_start, filter_end)
        with concurrent.futures.ProcessPoolExecutor() as executor:
            # warning: Breakpoints in the underlying method might not work when started from ipynb
            pa_from_csv_for_participant = partial(
                ParticipantsResultContainer.from_csv_for_participant,
                simulation_data_path,
                delimiter,
                simulation_end,
                grid_container,
            )
            participant_results = executor.map(
                pa_from_csv_for_participant,
                filter(
                    lambda x: x != SystemParticipantsEnum.FLEX_OPTIONS,
                    SystemParticipantsEnum.values(),
                ),
            )
            participant_result_map = {}
            for participant_result in participant_results:
                participant_result_map[
                    participant_result.entity_type
                ] = participant_result

        res = ParticipantsResultContainer(
            loads=participant_result_map[SystemParticipantsEnum.LOAD],
            fixed_feed_ins=participant_result_map[SystemParticipantsEnum.FIXED_FEED_IN],
            pvs=participant_result_map[SystemParticipantsEnum.PHOTOVOLTAIC_POWER_PLANT],
            wecs=participant_result_map[SystemParticipantsEnum.WIND_ENERGY_CONVERTER],
            storages=participant_result_map[SystemParticipantsEnum.STORAGE],
            ems=participant_result_map[SystemParticipantsEnum.ENERGY_MANAGEMENT],
            evcs=participant_result_map[SystemParticipantsEnum.EV_CHARGING_STATION],
            evs=participant_result_map[SystemParticipantsEnum.ELECTRIC_VEHICLE],
            hps=participant_result_map[SystemParticipantsEnum.HEAT_PUMP],
            flex=FlexOptionsResult.from_csv(
                SystemParticipantsEnum.FLEX_OPTIONS,
                simulation_data_path,
                delimiter,
                simulation_end,
            ),
        )
        return (
            res
            if not filter_start
            else res.filter_for_time_interval(filter_start, filter_end)
        )

    @staticmethod
    def from_csv_for_participant(
        simulation_data_path: str,
        delimiter: str,
        simulation_end: datetime,
        grid_container: Optional[GridContainer],
        participant: SystemParticipantsEnum,
    ):
        if grid_container:
            input_entities = grid_container.participants.get_participants(participant)
        else:
            input_entities = None
        if participant.has_soc():
            return ParticipantsWithSocResult.from_csv(
                participant,
                simulation_data_path,
                delimiter,
                simulation_end,
                input_entities,
            )
        else:
            return ParticipantsResult.from_csv(
                participant,
                simulation_data_path,
                delimiter,
                simulation_end,
                input_entities=input_entities,
            )

    def to_list(
        self, include_em: bool = True, include_flex=True, include_empty=True
    ) -> list[ParticipantsResult]:
        optional = []
        if include_em:
            optional.append(self.ems)
        if include_flex:
            optional.append(self.flex)
        required_participants = [
            self.loads,
            self.hps,
            self.fixed_feed_ins,
            self.wecs,
            self.storages,
            self.evcs,
            self.pvs,
        ]
        all_participants = required_participants + optional
        if not include_empty:
            all_participants = [
                participants
                for participants in all_participants
                if participants.entities
            ]
        return all_participants

    def subset(self, uuids):
        return ParticipantsResultContainer(
            self.ems.subset(uuids),
            self.loads.subset(uuids),
            self.fixed_feed_ins.subset(uuids),
            self.pvs.subset(uuids),
            self.wecs.subset(uuids),
            self.storages.subset(uuids),
            self.evcs.subset(uuids),
            self.evs.subset(uuids),
            self.hps.subset(uuids),
            self.flex.subset(uuids),
        )

    def get_participants(self, sp_type: SystemParticipantsEnum):
        if sp_type == SystemParticipantsEnum.ENERGY_MANAGEMENT:
            return self.ems
        elif sp_type == SystemParticipantsEnum.LOAD:
            return self.loads
        elif sp_type == SystemParticipantsEnum.FIXED_FEED_IN:
            return self.fixed_feed_ins
        elif sp_type == SystemParticipantsEnum.PHOTOVOLTAIC_POWER_PLANT:
            return self.pvs
        elif sp_type == SystemParticipantsEnum.WIND_ENERGY_CONVERTER:
            return self.wecs
        elif sp_type == SystemParticipantsEnum.STORAGE:
            return self.storages
        elif sp_type == SystemParticipantsEnum.EV_CHARGING_STATION:
            return self.evcs
        elif sp_type == SystemParticipantsEnum.HEAT_PUMP:
            return self.hps
        else:
            raise ValueError(
                f"No return value for system participant of type: {sp_type}"
            )

    @property
    def p(self) -> DataFrame:
        p_series = {
            participants.entity_type.value: participants.p_sum()
            for participants in self.to_list(include_flex=False)
        }
        return pd.DataFrame(p_series).sort_index().ffill().fillna(0)

    def p_sum(self) -> Series:
        return self.p.sum(axis=1).rename("p_sum")

    @property
    def q(self) -> DataFrame:
        q_series = {
            participants.entity_type.value: participants.q_sum()
            for participants in self.to_list(include_flex=False)
        }
        return pd.DataFrame(q_series).sort_index().ffill().fillna(0)

    def q_sum(self) -> Series:
        return self.q.sum(axis=1).rename("q_sum")

    def find_participant_result(self, uuid: str):
        for participants_res in self.to_list(include_flex=False):
            if uuid in participants_res:
                return participants_res.get(uuid)
        return ParticipantsResult.create_empty("None")

    def to_dict(
        self, include_empty: bool = True
    ) -> dict[SystemParticipantsEnum, ParticipantsResult]:
        if include_empty:
            return {res.entity_type: res for res in self.to_list()}
        else:
            return {res.entity_type: res for res in self.to_list() if res.entities}

    def energies(self) -> dict[SystemParticipantsEnum, float]:
        return {
            sp_type: res.energy()
            for sp_type, res in self.to_dict(include_empty=False).items()
            if sp_type != SystemParticipantsEnum.FLEX_OPTIONS
        }

    def load_and_generation_energies(self) -> dict[SystemParticipantsEnum, float]:
        return {
            sp_type: res.load_and_generation()
            for sp_type, res in self.to_dict(include_empty=False).items()
        }

    def sum(self) -> PQResult:
        participant_res = []
        for participant in self.to_list(include_em=False, include_flex=False):
            participant_res.append(participant.sum())
        return PQResult.sum(participant_res)

    def filter_by_date_time(self, time: Union[datetime, list[datetime]]):
        return ParticipantsResultContainer(
            self.ems.filter_by_date_time(time),
            self.loads.filter_by_date_time(time),
            self.fixed_feed_ins.filter_by_date_time(time),
            self.pvs.filter_by_date_time(time),
            self.wecs.filter_by_date_time(time),
            self.storages.filter_by_date_time(time),
            self.evcs.filter_by_date_time(time),
            self.evs.filter_by_date_time(time),
            self.hps.filter_by_date_time(time),
            self.flex.filter_by_date_time(time),
        )

    def filter_for_time_interval(self, start: datetime, end: datetime):
        return ParticipantsResultContainer(
            self.ems.filter_for_time_interval(start, end),
            self.loads.filter_for_time_interval(start, end),
            self.fixed_feed_ins.filter_for_time_interval(start, end),
            self.pvs.filter_for_time_interval(start, end),
            self.wecs.filter_for_time_interval(start, end),
            self.storages.filter_for_time_interval(start, end),
            self.evcs.filter_for_time_interval(start, end),
            self.evs.filter_for_time_interval(start, end),
            self.hps.filter_for_time_interval(start, end),
            self.flex.filter_for_time_interval(start, end),
        )



from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Union

from networkx import Graph

from psdm_analysis.models.entity import Entities
from psdm_analysis.models.input.connector.lines import Lines
from psdm_analysis.models.input.connector.switches import Switches
from psdm_analysis.models.input.connector.transformer import Transformers2W
from psdm_analysis.models.input.container.mixins import ContainerMixin
from psdm_analysis.models.input.container.participants_container import (
    SystemParticipantsContainer,
)
from psdm_analysis.models.input.node import Nodes
from psdm_analysis.models.primary_data import PrimaryData
from psdm_analysis.models.result.power import PQResult


@dataclass(frozen=True)
class RawGridContainer(ContainerMixin):
    nodes: Nodes
    lines: Lines
    transformers_2_w: Transformers2W
    switches: Switches

    def to_list(self, include_empty: bool = False) -> list[Entities]:
        grid_elements = [self.nodes, self.lines, self.transformers_2_w, self.switches]
        return grid_elements if include_empty else [e for e in grid_elements if e]

    @classmethod
    def from_csv(cls, path: str, delimiter: str) -> "RawGridContainer":
        nodes = Nodes.from_csv(path, delimiter)
        lines = Lines.from_csv(path, delimiter)
        transformers_2_w = Transformers2W.from_csv(path, delimiter)
        switches = Switches.from_csv(path, delimiter)
        return cls(
            nodes=nodes,
            lines=lines,
            transformers_2_w=transformers_2_w,
            switches=switches,
        )

    def get_branches(self) -> list[list[str]]:
        """
        Returns all branches, branching off from the slack node of the grid.
        The branches are returned as a list of lists, where each list contains the node uuids of the branch,
        starting at the slack node.

        Currently only works for single slack node and single voltage level grids.
        """

        slack_node = self.nodes.get_slack_nodes()
        if len(slack_node.data) != 1:
            raise ValueError("Currently only implemented for singular slack nodes.")
        transformers = self.transformers_2_w
        slack_transformers = Transformers2W(
            transformers.data[
                (transformers.node_a.isin(slack_node.uuids.to_list()))
                | (transformers.node_b.isin(slack_node.uuids.to_list()))
            ]
        )
        slack_connected_node = (
            set(slack_transformers.node_a)
            .union(slack_transformers.node_b)
            .difference(slack_node.uuids)
        )
        if len(slack_connected_node) > 1:
            raise ValueError(
                "There are multiple nodes connected to the slack node via a transformer."
            )
        elif len(slack_connected_node) == 0:
            raise ValueError("Did not find a slack node!")
        return self._find_branches(
            self.build_networkx_graph(), slack_connected_node.pop()
        )

    @staticmethod
    def _find_branches(G: Graph, start_node):
        visited = set()
        visited.add(start_node)
        branches = []

        def dfs(node, path):
            visited.add(node)
            path.append(node)

            for neighbor in G.neighbors(node):
                if neighbor not in visited:
                    dfs(neighbor, path)

        for neighbor in G.neighbors(start_node):
            if neighbor not in visited:
                path = []
                dfs(neighbor, path)
                branches.append(path)

        return branches

    def build_networkx_graph(self) -> Graph:
        graph = Graph()
        closed_switches = self.switches.get_closed()
        line_data_dicts = self.lines.data.apply(
            lambda row: {"length": row["length"]}, axis=1
        )

        graph.add_nodes_from(self.nodes.uuids)
        graph.add_edges_from(zip(self.lines.node_a, self.lines.node_b, line_data_dicts))
        graph.add_edges_from(zip(closed_switches.node_a, closed_switches.node_b))
        return graph


@dataclass(frozen=True)
class GridContainer(ContainerMixin):
    raw_grid: RawGridContainer
    # todo: we keep the participant containers effectively twice with the mapping
    participants: SystemParticipantsContainer
    primary_data: PrimaryData
    node_participants_map: Dict[str, SystemParticipantsContainer]

    @classmethod
    def from_csv(cls, path: str, delimiter: str, primary_data_delimiter: str = None):
        if not primary_data_delimiter:
            primary_data_delimiter = delimiter
        raw_grid = RawGridContainer.from_csv(path, delimiter)
        participants = SystemParticipantsContainer.from_csv(path, delimiter)
        node_participants_map = {
            uuid: participants.filter_by_node(uuid) for uuid in raw_grid.nodes.uuids
        }
        primary_data = PrimaryData.from_csv(path, primary_data_delimiter)
        return cls(raw_grid, participants, primary_data, node_participants_map)

    def to_list(self, include_empty: bool = False, include_primary_data: bool = False):
        grid = [self.raw_grid, self.participants]
        return grid if not include_primary_data else grid + [self.primary_data]

    def get_nodal_primary_data(self):
        time_series = []
        nodal_primary_data = dict()
        for node, participants_container in self.node_participants_map.items():
            participants_uuids = participants_container.uuids().tolist()
            node_primary_data = self.primary_data.get_for_participants(
                participants_uuids
            )
            time_series.extend(node_primary_data)
            node_primary_data_agg = PQResult.sum(node_primary_data)
            nodal_primary_data[node] = node_primary_data_agg
        return nodal_primary_data

    def get_nodal_sp_count_and_power(self):
        data = {}
        for node_uuid, sps in self.node_participants_map.items():
            nodal_data = {}
            for sp in sps.to_list(include_empty=False):
                sp_id = sp.get_enum().value
                count = len(sp.data)
                data_str = f"Count: {count}"
                # check if sp has a property named s_rated
                if hasattr(sp, "s_rated"):
                    s_rated = round(sp.s_rated.sum(), 2)
                    data_str += f", Rated Power: {s_rated} kw"
                nodal_data[sp_id] = data_str
            data[node_uuid] = nodal_data
        return data

    def filter_by_date_time(self, time: Union[datetime, list[datetime]]):
        return GridContainer(
            raw_grid=self.raw_grid,
            participants=self.participants,
            primary_data=self.primary_data.filter_by_date_time(time),
            node_participants_map=self.node_participants_map,
        )



from dataclasses import dataclass

from psdm_analysis.models.input.container.mixins import SpTypeMixin
from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.input.participant.participant import (
    SystemParticipantsWithCapacity,
)


@dataclass(frozen=True)
class ElectricVehicles(SystemParticipantsWithCapacity, SpTypeMixin):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.ELECTRIC_VEHICLE

    @staticmethod
    def capacity_attribute() -> str:
        return "e_storage"

    @property
    def e_storage(self):
        return self.data["e_storage"]

    @property
    def e_cons(self):
        return self.data["e_cons"]

    @staticmethod
    def entity_attributes() -> list[str]:
        return []

    @staticmethod
    def type_attributes() -> list[str]:
        return SpTypeMixin.type_attributes() + [
            "e_storage",
            "e_cons",
        ]



import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns

from psdm_analysis.models.result.participant.participants_res_container import (
    ParticipantsResultContainer,
)
from psdm_analysis.models.result.power import PQResult
from psdm_analysis.plots.utils import (
    FIGSIZE,
    FIGSIZE_WIDE,
    LOAD_COLOR,
    PV_COLOR,
    TITLE_FONT_SIZE,
    get_label_and_color,
)

sns.set_style("whitegrid")


def daily_usage(res: PQResult, device_power_mw):
    fig, ax = plt.subplots(figsize=FIGSIZE_WIDE)
    usage_hours = res.daily_usage(device_power_mw)
    label, color = get_label_and_color(res.type)
    ax.bar(usage_hours.index, usage_hours, color=color)
    ax.set_ylabel("Daily usage in hours")
    ax.set_xlabel("Day of year")


def plot_load_and_generation(participant_res: ParticipantsResultContainer):
    lg_dict = participant_res.load_and_generation_energies()
    keys = []
    load = []
    generation = []
    for key, load_generation in lg_dict.items():
        keys.append(key)
        load.append(load_generation[0])
        generation.append(load_generation[1])
    keys = [key.value for key in lg_dict.keys()]
    fig, axs = plt.subplots(figsize=FIGSIZE, ncols=2, sharey=True)
    plt.subplots_adjust(wspace=0)

    x_lim = max(max(load), abs(min(generation))) * 1.2

    bars_generation = axs[0].barh(
        keys,
        generation,
        align="center",
        color=PV_COLOR,
        edgecolor=PV_COLOR,
        alpha=0.7,
        zorder=10,
    )
    axs[0].set_title("Generation", fontsize=TITLE_FONT_SIZE, pad=15)
    axs[0].set_xlim(left=-x_lim)
    axs[0].bar_label(
        bars_generation,
        padding=4,
        labels=[f"{x:,.0f}" for x in bars_generation.datavalues],
    )
    axs[0].get_xaxis().set_major_formatter(
        mpl.ticker.FuncFormatter(lambda x, p: format(int(x), ","))
    )

    bars_load = axs[1].barh(
        keys, load, align="center", color=LOAD_COLOR, alpha=0.7, zorder=10
    )
    axs[1].bar_label(
        bars_load, padding=4, labels=[f"{x:,.0f}" for x in bars_load.datavalues]
    )
    axs[1].set_title("Load", fontsize=TITLE_FONT_SIZE, pad=15)
    axs[1].set_xlim(right=x_lim)
    axs[1].get_xaxis().set_major_formatter(
        mpl.ticker.FuncFormatter(lambda x, p: format(int(x), ","))
    )
    # this should work since mpl 3.4 but does not for some reason

    # axs[1].bar_label(bars_load)
    fig.tight_layout()
    plt.show()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, List

from pandas import Series

from psdm_analysis.models.result.grid.connector import ConnectorResult, ConnectorsResult

if TYPE_CHECKING:
    from psdm_analysis.models.grid_with_results import GridWithResults


@dataclass(frozen=True)
class Transformer2WResult(ConnectorResult):
    @staticmethod
    def attributes() -> List[str]:
        return ConnectorResult.attributes() + ["tap_pos"]

    @property
    def tap_pos(self) -> Series:
        return self.data["tap_pos"]

    def calc_rated_power_gwr(self, gwr: GridWithResults, side="hv", absolute=True):
        if side == "hv":
            node = "a"
        elif side == "lv":
            node = "b"
        else:
            raise ValueError('Side has to be either "hv" or "lv"')
        transformer = gwr.grid.raw_grid.transformers_2_w.subset(self.input_model)
        node_uuid = transformer.node_a if side == "hv" else transformer.node_b
        node_uuid = node_uuid.to_list()[0]
        v_rated = gwr.grid.raw_grid.nodes.v_rated.loc[node_uuid]
        node_res = gwr.results.nodes[node_uuid]
        return self.calc_rated_power(node_res, v_rated, node, absolute)

    def calc_transformer_utilisation(
        self, gwr: GridWithResults, side="hv", absolute=True
    ):
        transformer = gwr.grid.raw_grid.transformers_2_w.subset(self.input_model)
        # todo: adjust to real prefix
        s_rated = transformer.s_rated.to_list()[0]
        return self.calc_rated_power_gwr(gwr, side, absolute) / s_rated


@dataclass(frozen=True)
class Transformers2WResult(ConnectorsResult):
    entities: dict[str, Transformer2WResult]



from functools import partial
from typing import Union

import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

from psdm_analysis.models.input.participant.participant import (
    SystemParticipantsWithCapacity,
)
from psdm_analysis.models.result.participant.participant import (
    ParticipantsResult,
    ParticipantsWithSocResult,
)
from psdm_analysis.models.result.participant.participants_res_container import (
    ParticipantsResultContainer,
)
from psdm_analysis.models.result.power import PQResult, PQWithSocResult
from psdm_analysis.plots.common.line_plot import ax_plot_time_series
from psdm_analysis.plots.common.utils import (
    FIGSIZE,
    FILL_ALPHA,
    LOAD_COLOR,
    PV_COLOR,
    add_to_kwargs_if_not_exist,
    get_label_and_color,
    get_label_and_color_dict,
    legend_with_distinct_labels,
    plot_resample,
    set_date_format_and_label,
    set_title,
)

sns.set_style("whitegrid")


def plot_comparison(
    res_a: PQResult,
    res_b: PQResult,
    label_a: str,
    label_b: str,
    title: str,
    resolution: str,
    hourly_mean=False,
    flex_signal: PQResult = None,
):
    subplot_count = 2 if flex_signal is None else 3
    fig, axs = plt.subplots(
        subplot_count, 1, figsize=FIGSIZE, sharex=True, sharey=False
    )
    ax_plot_active_power(
        axs[0], res_a, resolution, hourly_mean, label=label_a, color=LOAD_COLOR
    )
    ax_plot_active_power(
        axs[0], res_b, resolution, hourly_mean, label=label_b, color=PV_COLOR
    )
    residual = res_a - res_b
    ax_plot_active_power(
        axs[1], residual, resolution, hourly_mean, label="Residual Load"
    )
    if flex_signal:
        ax_plot_active_power(
            axs[2], flex_signal, resolution, hourly_mean, label="Flex Signal"
        )
    fig.suptitle(title)
    [ax.legend() for ax in axs]
    return axs, fig


def plot_em(
    em_participant_results: ParticipantsResultContainer,
    resolution: str,
    hourly_mean: bool = False,
):
    em_uuid = list(em_participant_results.ems.participants.keys())[0]

    title = f"Household: {em_uuid}"

    fig, axs = plt.subplots(2, 1, figsize=(10, 5), sharex=True, sharey=True)
    axs[0].set_title(title)

    ax_plot_participants(
        axs[0], em_participant_results, resolution, hourly_mean=hourly_mean, stack=True
    )
    ax_plot_active_power(
        axs[1],
        em_participant_results.ems.sum(),
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=True,
    )
    return fig


def plot_aggregated_load_and_generation(
    participants: Union[ParticipantsResultContainer, list[PQResult]],
    title: str,
    resolution: str,
    hourly_mean: bool,
    with_residual=True,
):
    if with_residual:
        fig, axs = plt.subplots(2, 1, figsize=(10, 5), sharex=True, sharey=True)
        pq_results = _get_pq_results_from_union(participants)
        all_load, all_generation = zip(
            *[pq_result.divide_load_generation() for pq_result in pq_results]
        )
        agg_load = PQResult.sum(all_load)
        agg_generation = PQResult.sum(all_generation)
        ax_plot_active_power(
            axs[0],
            agg_load,
            resolution,
            hourly_mean,
            fill_from_index=True,
            color=LOAD_COLOR,
            label="Aggregated Load",
        )
        ax_plot_active_power(
            axs[0],
            agg_generation,
            resolution,
            hourly_mean,
            fill_from_index=True,
            color=PV_COLOR,
            label="Aggregated Generation",
        )
        ax_plot_active_power(
            axs[1],
            PQResult.sum([agg_load, agg_generation]),
            resolution,
            hourly_mean,
            fill_from_index=True,
            color=LOAD_COLOR,
            label="Residual Load",
        )
        set_title(axs[0], title)
        axs[0].legend(loc=7)
        axs[1].legend()
        return fig, axs


def plot_all_participants(
    participants: Union[ParticipantsResultContainer, list[PQResult]],
    title: str,
    resolution: str,
    hourly_mean: bool,
    stack=False,
    with_residual=False,
    **kwargs,
):
    if with_residual:
        fig, axs = plt.subplots(2, 1, figsize=(10, 5), sharex=True, sharey=True)
        ax_plot_participants(
            axs[0], participants, resolution, hourly_mean, stack, **kwargs
        )
        participants_sum = (
            participants.sum()
            # here I would usually check if participants is a ParticipantsResultContainer
            # but for some weird reason this returns False every time
            if not isinstance(participants, list)
            else PQResult.sum(participants)
        )
        ax_plot_active_power(
            axs[1],
            participants_sum,
            resolution,
            hourly_mean,
            fill_from_index=True,
            **kwargs,
        )
        set_title(axs[0], title)
    else:
        fig, axs = plt.subplots(figsize=FIGSIZE)
        ax_plot_participants(
            axs, participants, resolution, hourly_mean, stack, **kwargs
        )
        set_title(axs, title)
    return fig, axs


def _get_pq_results_from_union(
    participants: Union[ParticipantsResultContainer, list[PQResult]]
) -> list[PQResult]:
    return (
        participants
        if isinstance(participants, list)
        else [
            participant.sum()
            for participant in participants.to_list(
                include_em=False, include_flex=False
            )
            if participant
        ]
    )


def ax_plot_participants(
    ax: Axes,
    participants: Union[ParticipantsResultContainer, list[PQResult]],
    resolution: str,
    hourly_mean: bool = False,
    stack=False,
    **kwargs,
):
    pq_results = _get_pq_results_from_union(participants)
    if stack:
        plot_kwargs = [kwargs for _ in range(len(pq_results))]
        ax_plot_stacked_pq(ax, pq_results, resolution, hourly_mean, plot_kwargs)
    else:
        [
            ax_plot_active_power(ax, pq_result, resolution, hourly_mean, **kwargs)
            for pq_result in pq_results
        ]
    legend_with_distinct_labels(ax)


def ax_plot_stacked_pq(
    ax: Axes,
    results: list[PQResult],
    resolution: str,
    hourly_mean: bool = False,
    plot_kwargs: list[dict] = None,
):
    residual_load, residual_generation = results[0].divide_load_generation()

    plot_partial = partial(
        ax_plot_active_power,
        ax=ax,
        resolution=resolution,
        hourly_mean=hourly_mean,
    )

    plot_partial(
        res=residual_load, fill_from_index=True, **plot_kwargs[0] if plot_kwargs else {}
    )
    plot_partial(
        res=residual_generation,
        fill_from_index=True,
        **plot_kwargs[0] if plot_kwargs else {},
    )

    for idx, res in enumerate(results[1:]):
        load, generation = res.divide_load_generation()

        if load.p.sum() > 0:
            load_sum = PQResult(
                res.type, "", "", PQResult.sum([load, residual_load]).data
            )
            plot_partial(
                res=load_sum,
                fill_between=residual_load.p,
                **plot_kwargs[idx + 1] if plot_kwargs else {},
            )
            residual_load = load_sum

        if generation.p.sum() < 0:
            generation_sum = PQResult(
                res.type, "", "", PQResult.sum([generation, residual_generation]).data
            )
            plot_partial(
                res=generation_sum,
                fill_between=residual_generation.p,
                **plot_kwargs[idx + 1] if plot_kwargs else {},
            )
            residual_generation = generation_sum
    handles, labels = ax.get_legend_handles_labels()
    by_label = dict(zip(labels, handles))
    ax.legend(by_label.values(), by_label.keys())


def plot_participants_sum(
    res: ParticipantsResult,
    title: str,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index: bool = True,
    **kwargs,
):
    if not isinstance(res, ParticipantsResult):
        raise TypeError(
            "Data must be of type ParticipantsResult but is {}".format(type(res))
        )
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_active_power(
        ax, res.sum(), resolution, hourly_mean, fill_from_index, **kwargs
    )
    set_title(ax, title)
    return fig


def plot_participants_with_soc_sum(
    res: ParticipantsWithSocResult,
    input: SystemParticipantsWithCapacity,
    title: str,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index: bool = True,
    **kwargs,
):
    fig, axs = plt.subplots(2, 1, figsize=FIGSIZE, sharex=True, sharey=False)
    axs[0].set_title(title)
    sum = ParticipantsWithSocResult.sum_with_soc(res, input)
    ax_plot_active_power(
        axs[0],
        sum,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        **kwargs,
    )
    ax_plot_soc(axs[1], sum, resolution, hourly_mean, fill_from_index, **kwargs)
    return fig


def plot_active_power_with_soc(
    res: PQWithSocResult,
    title: str,
    resolution: str,
    hourly_mean=False,
    fill_from_index=False,
    **kwargs,
):
    fig, axs = plt.subplots(2, 1, figsize=FIGSIZE, sharex=True, sharey=False)
    axs[0].set_title(title)
    ax_plot_active_power(
        axs[0],
        res,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        set_x_label=False,
        **kwargs,
    )
    ax_plot_soc(axs[1], res, resolution, hourly_mean, fill_from_index, **kwargs)
    return fig


def plot_active_power(
    res: PQResult,
    title: str,
    resolution: str,
    hourly_mean: bool,
    fill_from_index=True,
    **kwargs,
):
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax.set_ylabel("Power in MW")
    ax_plot_active_power(
        ax, res, resolution, hourly_mean, fill_from_index=fill_from_index, **kwargs
    )
    set_title(ax, title)
    return fig, ax


def ax_plot_active_power(
    ax: Axes,
    res: PQResult,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.p) == 0:
        raise ValueError("Active power time series is empty. No data to plot")

    ax = ax_plot_time_series(
        ax,
        res.p,
        res.type,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Power in MW")


def ax_plot_soc(
    ax: Axes,
    res: PQWithSocResult,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index=False,
    **kwargs,
):
    label, color = get_label_and_color(res.type)
    set_date_format_and_label(ax, resolution)
    ax.set_ylabel("SOC in percent")
    soc = plot_resample(res.soc(), hourly_mean)
    ax.plot(soc, color=color, label=label, **kwargs)
    ax.set_ylim(bottom=0, top=100)
    if fill_from_index:
        ax.fill_between(soc.index, soc, alpha=FILL_ALPHA, color=color)


def plot_sorted_annual_load_duration(
    res: PQResult, s_rated_mw: float = None, fill_from_index=False, **kwargs
):
    args = get_label_and_color_dict(res.type)
    kwargs = add_to_kwargs_if_not_exist(kwargs, args)
    fig, ax = plt.subplots(figsize=FIGSIZE)
    annual_duartion_series = res.annual_duration_series()
    plt.plot(annual_duartion_series, **kwargs)
    set_title(ax, f"{kwargs['label']} Annual Load Duration Curve")
    ax.set_xlabel("Duration in hours")
    ax.set_ylabel("Rated power")
    if fill_from_index:
        ax.fill_between(
            annual_duartion_series.index,
            annual_duartion_series,
            alpha=FILL_ALPHA,
            color=kwargs["color"],
        )
    if s_rated_mw:
        # todo: implement right hand y-axis for displaying s_rated percentage
        pass
    return fig, ax



from typing import Tuple

import pandas as pd
from pandas import DataFrame, Series


def duration_weighted_series(series: Series):
    series.sort_index(inplace=True)
    values = series[:-1].reset_index(drop=True)
    time = (
        (series.index[1::] - series.index[:-1])
        .to_series()
        .apply(lambda x: x.total_seconds() / 3600)
        .reset_index(drop=True)
    )
    return pd.concat([values.rename("values"), time.rename("time")], axis=1)


def weighted_series_sum(weighted_series: DataFrame) -> float:
    return (weighted_series["values"] * weighted_series["time"]).sum()


def duration_weighted_sum(series: Series) -> float:
    weighted_series = duration_weighted_series(series)
    return weighted_series_sum(weighted_series)


def add_series(this: Series, that: Series, name: str) -> Series:
    # todo: Is there a more performant implementation?
    return (
        this.to_frame()
        .join(that.rename("other"), how="outer")
        .fillna(method="ffill")
        .sum(axis=1)
        .rename(name)
    )


def join_series(series: list[Series]) -> DataFrame:
    first_series = series[0]
    df = first_series.rename(first_series.name).to_frame()
    for series in series[1:]:
        df = df.join(series.rename(series.name), how="outer").fillna(method="ffill")
    return df


def hourly_mean_resample(series: Series) -> Series:
    return series.resample("60s").ffill().resample("1h").mean()


def load_and_generation(p_ts: Series) -> Tuple[float, float]:
    weighted_series = duration_weighted_series(p_ts)
    load_filter = weighted_series["values"] > 0
    weighted_load = weighted_series[load_filter]
    weighted_generation = weighted_series[~load_filter]
    return (
        weighted_series_sum(weighted_load),
        weighted_series_sum(weighted_generation),
    )


def divide_positive_negative(series: Series):
    positive = series.copy(deep=True)
    positive[positive < 0] = 0
    negative = series.copy(deep=True)
    negative[negative > 0] = 0
    return positive, negative


def p_to_pq_frame(p: Series):
    if not p.name == "p":
        p = p.rename("p")
    data = p.to_frame()
    data["q"] = 0
    return data



import logging
from abc import ABC, abstractmethod

from psdm_analysis.io.utils import df_to_csv
from psdm_analysis.models.entity import Entities
from psdm_analysis.models.input.enums import EntitiesEnum
from psdm_analysis.models.input.participant.participant import SystemParticipants


class ContainerMixin(ABC):
    @abstractmethod
    def to_list(self, include_empty: bool = False) -> list:
        pass

    def to_csv(self, path: str, delimiter: str):
        [e.to_csv(path, delimiter) for e in self.to_list()]


class HasTypeMixin(ABC):
    """
    Mixin for entities that have a separate type within the psdm data model.
    This should only be extended as an addition to the Entities class.
    """

    @staticmethod
    @abstractmethod
    def get_enum() -> EntitiesEnum:
        pass

    @property
    def type_id(self):
        return self.data["type_id"]

    @property
    def type_uuid(self):
        return self.data["type_uuid"]

    def to_csv(self, path: str, delimiter: str = ","):
        # persist entity_input.csv
        all_entity_attributes = self.attributes(include_type_attrs=False)
        all_entity_attributes.append("type_uuid"), all_entity_attributes.remove(
            "uuid"
        )  # uuid is set as index
        entity_data = self.data[all_entity_attributes].rename(
            columns={"type_uuid": "type"}
        )
        df_to_csv(
            entity_data, path, self.get_enum().get_csv_input_file_name(), delimiter
        )

        # persist entity_type_input.csv
        type_data = (
            self.data[self.type_attributes()]
            .set_index("type_uuid", drop=True)
            .rename(columns={"type_id": "id"})
        )
        df_to_csv(
            type_data.drop_duplicates(),
            path,
            self.get_enum().get_type_file_name(),
            delimiter,
        )

    @classmethod
    def attributes(cls, include_type_attrs: bool = True) -> list[str]:
        base_attributes = []
        # check if cls is a subclass of SystemParticipants
        if issubclass(cls, SystemParticipants):
            base_attributes += SystemParticipants.attributes()
        elif issubclass(cls, Entities):
            base_attributes += Entities.attributes()
        else:
            logging.warning(
                "We expect the type that extends this mixin to be a subclass of a type with attributes!"
            )
        all_entity_attributes = base_attributes + cls.entity_attributes()
        return (
            all_entity_attributes + cls.type_attributes()
            if include_type_attrs
            else all_entity_attributes
        )

    @staticmethod
    @abstractmethod
    def entity_attributes() -> list[str]:
        pass

    @staticmethod
    def type_attributes() -> list[str]:
        return ["type_uuid", "type_id"]


class SpTypeMixin(HasTypeMixin):
    @property
    def capex(self):
        return self.data["capex"]

    @property
    def opex(self):
        return self.data["opex"]

    @property
    def s_rated(self):
        return self.data["s_rated"]

    @property
    def cos_phi_rated(self):
        return self.data["cos_phi_rated"]

    @staticmethod
    @abstractmethod
    def entity_attributes() -> list[str]:
        pass

    @staticmethod
    def type_attributes() -> list[str]:
        return HasTypeMixin.type_attributes() + [
            "capex",
            "opex",
            "s_rated",
            "cos_phi_rated",
        ]



import numpy as np
from pandas import Series


def rmse_by_iloc(predictions: Series, targets: Series):
    """
    Calculates the RMSE between two series. Resets the index to calculate by their index (iloc)
    :param predictions: the predictions
    :param targets: the targets to aim for
    :return: the root mean squared error
    """
    if len(predictions) != len(targets):
        raise ValueError("Predictions and targets should have the same length!")
    return np.sqrt(
        (
            (predictions.reset_index(drop=True) - targets.reset_index(drop=True)) ** 2
        ).mean()
    )



from dataclasses import dataclass
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from pandas import DataFrame, Series

from psdm_analysis.models.entity import ResultEntities
from psdm_analysis.models.input.enums import RawGridElementsEnum
from psdm_analysis.models.input.node import Nodes
from psdm_analysis.models.result.participant.dict import ResultDict


@dataclass(frozen=True)
class NodeResult(ResultEntities):
    def __eq__(self, other):
        if not isinstance(other, NodeResult):
            return False
        return (
            (self.input_model == other.input_model)
            & (self.name == other.name)
            & (self.data.equals(other.data))
        )

    @staticmethod
    def attributes() -> List[str]:
        return ["v_ang", "v_mag"]

    # todo: fix me
    @staticmethod
    def build_from_nominal_data(
        uuid: str,
        name: Optional[str],
        data: DataFrame,
        rated_voltage: float,
    ) -> "NodeResult":
        data["v_mag"] = data["v_mag"].divide(rated_voltage)
        return NodeResult(RawGridElementsEnum.NODE, uuid, name, data)

    @property
    def v_mag(self) -> Series:  # in Ampere
        return self.data["v_mag"]

    @property
    def v_ang(self) -> Series:
        return self.data["v_ang"]

    def v_complex(self, v_rated_kv_src: Union[float, Nodes]) -> Series:
        v_rated_kv = (
            v_rated_kv_src
            if isinstance(v_rated_kv_src, float)
            else v_rated_kv_src.subset(self.input_model).v_rated.iloc[0]
        )
        return (self.v_mag * v_rated_kv) * np.exp(1j * np.radians(self.v_ang))


@dataclass(frozen=True)
class NodesResult(ResultDict):
    entities: dict[str, NodeResult]

    @property
    def v_mag(self) -> Optional[DataFrame]:
        if not self.entities:
            return None
        return pd.concat(
            [
                node_res.v_mag.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )

    @property
    def v_ang(self) -> Optional[DataFrame]:
        if not self.entities:
            return None
        return pd.concat(
            [
                node_res.v_ang.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )

    def v_mag_describe(self) -> DataFrame:
        return self._describe(self.v_mag)

    def v_ang_describe(self) -> DataFrame:
        return self._describe(self.v_ang)

    @staticmethod
    def _describe(data: DataFrame):
        return data.describe().transpose()



from dataclasses import dataclass

from psdm_analysis.models.input.container.mixins import SpTypeMixin
from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class BiomassPlants(SystemParticipants, SpTypeMixin):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.BIOMASS_PLANT

    @property
    def market_reaction(self):
        return self.data["market_reaction"]

    @property
    def cost_controlled(self):
        return self.data["cost_controlled"]

    @property
    def feed_in_tariff(self):
        return self.data["feed_in_tariff"]

    @property
    def active_power_gradient(self):
        return self.data["active_power_gradient"]

    @property
    def eta_conv(self):
        return self.data["eta_conv"]

    @staticmethod
    def entity_attributes() -> list[str]:
        return SystemParticipants.attributes() + [
            "type",
            "market_reaction",
            "cost_controlled",
            "feed_in_tariff",
            "type_id",
            "active_power_gradient",
        ]

    @staticmethod
    def type_attributes() -> list[str]:
        return SpTypeMixin.type_attributes() + [
            "active_power_gradient",
            "eta_conv",
        ]



from dataclasses import dataclass
from typing import List

from pandas import DataFrame, Series

from psdm_analysis.models.input.connector.connector import Connector
from psdm_analysis.models.input.container.mixins import HasTypeMixin
from psdm_analysis.models.input.enums import RawGridElementsEnum


@dataclass(frozen=True)
class Lines(Connector, HasTypeMixin):
    data: DataFrame

    @staticmethod
    def get_enum() -> RawGridElementsEnum:
        return RawGridElementsEnum.LINE

    @property
    def length(self) -> Series:
        return self.data["length"]

    @property
    def geo_position(self) -> Series:
        return self.data["geo_position"]

    @property
    def olm_characteristic(self) -> Series:
        return self.data["olm_characteristic"]

    @property
    def v_rated(self) -> Series:
        return self.data["v_rated"]

    @property
    def r(self) -> Series:
        return self.data["r"]

    @property
    def x(self) -> Series:
        return self.data["x"]

    @property
    def b(self) -> Series:
        return self.data["b"]

    @property
    def i_max(self) -> Series:
        return self.data["i_max"]

    @staticmethod
    def entity_attributes() -> List[str]:
        return [
            "node_a",
            "node_b",
            "length",
            "geo_position",
            "olm_characteristic",
            "parallel_devices",
        ]

    @staticmethod
    def type_attributes() -> List[str]:
        return HasTypeMixin.attributes() + ["r", "x", "b", "i_max", "v_rated"]

    def aggregated_line_length(self) -> float:
        return self.data["length"].sum()

    def relative_line_length(self) -> float:
        return self.data["length"] / len(self.data)

    def find_lines_by_nodes(self, node_uuids):
        return Lines(
            self.data[(self.node_a.isin(node_uuids)) | (self.node_b.isin(node_uuids))]
        )



from pandas import Series

from psdm_analysis.models.input.container.mixins import SpTypeMixin
from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.input.participant.participant import (
    SystemParticipantsWithCapacity,
)


class Storages(SystemParticipantsWithCapacity, SpTypeMixin):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.STORAGE

    @property
    def behaviour(self) -> Series:
        return self.data["behaviour"]

    @staticmethod
    def capacity_attribute() -> str:
        return "e_storage"

    @property
    def p_max(self) -> Series:
        return self.data["p_max"]

    @property
    def active_power_gradient(self) -> Series:
        return self.data["active_power_gradient"]

    @property
    def eta(self) -> Series:
        return self.data["eta"]

    @property
    def dod(self) -> Series:
        return self.data["dod"]

    @property
    def life_time(self) -> Series:
        return self.data["life_time"]

    @property
    def life_cycle(self) -> Series:
        return self.data["life_cycle"]

    @staticmethod
    def entity_attributes() -> [str]:
        return [
            "behaviour",
        ]

    @staticmethod
    def type_attributes() -> [str]:
        return SpTypeMixin.type_attributes() + [
            "p_max",
            "active_power_gradient",
            "e_storage",
            "eta",
            "dod",
            "life_time",
            "life_cycle",
        ]



from abc import ABC, abstractmethod
from dataclasses import dataclass

from psdm_analysis.models.entity import Entities
from psdm_analysis.models.input.node import Nodes


@dataclass(frozen=True)
class SystemParticipants(Entities, ABC):
    @staticmethod
    def attributes():
        return Entities.attributes() + ["node", "q_characteristics"]

    def nodes(self):
        return self.data["node"]

    def q_characteristics(self):
        return self.data["q_characteristic"]

    def subset(self, uuids: list[str]):
        data = self.data.loc[self.data.index.intersection(uuids)]
        return type(self)(data)

    def insert_node_id_columns(self, nodes: Nodes) -> None:
        index_to_id = nodes.ids.to_dict()

        if "node_id" not in self.data.columns:
            self.data.insert(self.data.columns.get_loc("node") + 1, "node_id", None)
        self.data["node_id"] = self.nodes().map(index_to_id)


@dataclass(frozen=True)
class SystemParticipantsWithCapacity(SystemParticipants):
    @staticmethod
    @abstractmethod
    def capacity_attribute() -> str:
        pass

    @staticmethod
    def attributes():
        return Entities.attributes() + ["node", "capacity"]

    def capacity(self):
        return self.data[self.capacity_attribute()]



from dataclasses import dataclass

from pandas import DataFrame

from psdm_analysis.models.input.connector.connector import Connector
from psdm_analysis.models.input.enums import EntitiesEnum, RawGridElementsEnum


@dataclass(frozen=True)
class Switches(Connector):
    data: DataFrame

    @staticmethod
    def attributes() -> list[str]:
        return Connector.attributes() + ["closed"]

    @staticmethod
    def get_enum() -> EntitiesEnum:
        return RawGridElementsEnum.SWITCH

    @property
    def closed(self):
        return self.data["closed"]

    def get_closed(self) -> "Switches":
        return self.subset(self.data.query("closed == True").index)

    def get_opened(self) -> "Switches":
        return self.subset(self.data.query("closed == False").index)



import os.path
import uuid
from dataclasses import dataclass
from typing import Dict, Optional

import pandas as pd
from pandas import DataFrame, Series

from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.input.participant.participant import (
    SystemParticipantsWithCapacity,
)
from psdm_analysis.models.result.participant.dict import ResultDict
from psdm_analysis.models.result.power import PQResult, PQWithSocResult
from psdm_analysis.processing.dataframe import join_dataframes


@dataclass(frozen=True)
class ParticipantsResult(ResultDict):
    entities: Dict[str, PQResult]

    def to_csv(self, path: str, resample_rate: str = None):
        file_name = self.entity_type.get_csv_result_file_name()

        def resample(data: DataFrame, input_model: str):
            data = (
                data.resample("60s").ffill().resample(resample_rate).mean()
                if resample_rate
                else data
            )
            data["uuid"] = data.apply(lambda _: uuid.uuid4(), axis=1)
            data["input_model"] = input_model
            return data

        resampled = [
            resample(participant.data, input_model)
            for input_model, participant in self.entities.items()
        ]
        df = join_dataframes(resampled)
        df.to_csv(os.path.join(path, file_name))

    def nodal_results(self, node_uuid: str) -> "ParticipantsResult":
        pass

    def get(self, key: str) -> PQResult:
        return self.entities[key]

    def subset(self, uuids):
        return type(self)(
            self.entity_type,
            {uuid: self.entities[uuid] for uuid in self.entities.keys() & uuids},
        )

    @property
    def p(self) -> Optional[DataFrame]:
        if not self.entities.values():
            return None
        return (
            pd.DataFrame({p_uuid: res.p for p_uuid, res in self.entities.items()})
            .fillna(method="ffill")
            .sort_index()
        )

    def sum(self) -> PQResult:
        return PQResult.sum(list(self.entities.values()))

    def p_sum(self) -> Series:
        if not self.entities:
            return Series(dtype=float)
        return self.p.fillna(method="ffill").sum(axis=1).rename("p_sum")

    @property
    def q(self) -> Optional[DataFrame]:
        if not self.entities.values():
            return None
        return (
            pd.DataFrame({p_uuid: res.q for p_uuid, res in self.entities.items()})
            .fillna(method="ffill")
            .sort_index()
        )

    def q_sum(self):
        if not self.entities:
            return Series(dtype=float)
        return self.q.fillna(method="ffill").sum(axis=1).rename("q_sum")

    def energy(self):
        # todo make concurrent
        sum = 0
        for participant in self.entities.values():
            sum += participant.energy()
        return sum

    def load_and_generation(self):
        return self.sum().load_and_generation()


@dataclass(frozen=True)
class ParticipantsWithSocResult(ParticipantsResult):
    entity_type: SystemParticipantsEnum
    entities: Dict[str, PQWithSocResult]

    def sum_with_soc(self, inputs: SystemParticipantsWithCapacity) -> PQWithSocResult:
        if not self.entities:
            return PQWithSocResult.create_empty(self.entity_type, "", "")
        capacity_participant = []
        for participant_uuid, res in self.entities.items():
            capacity = inputs.get(participant_uuid)[inputs.capacity_attribute()]
            capacity_participant.append((capacity, res))
        return PQWithSocResult.sum_with_soc(capacity_participant)



from dataclasses import dataclass

from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class EvChargingStations(SystemParticipants):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.EV_CHARGING_STATION

    @property
    def charging_points(self):
        return self.data["charging_points"]

    @property
    def location_types(self):
        return self.data["location_type"]

    @property
    def s_rated(self):
        return self.data["s_rated"]

    @property
    def electric_current_type(self):
        return self.data["electric_current_type"]

    def get_public_evcs(self):
        return self.data[self.location_types.isin(["CUSTOMER_PARKING", "WORK"])]

    def get_home_evcs(self):
        return self.data[self.location_types.isin(["HOME"])]



from dataclasses import dataclass
from typing import List

from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class Loads(SystemParticipants):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.LOAD

    @property
    def s_rated(self):
        return self.data["s_rated"]

    @property
    def standard_load_profile(self):
        return self.data["standard_load_profile"]

    @property
    def e_cons_annual(self):
        return self.data["e_cons_annual"]

    @property
    def cosphi_rated(self):
        return self.data["cosphi_rated"]

    @staticmethod
    def attributes() -> List[str]:
        return SystemParticipants.attributes() + [
            "standard_load_profile",
            "e_cons_annual",
            "s_rated",
            "cosphi_rated",
        ]



import logging
from typing import Optional, Union

import matplotlib.pyplot as plt
from matplotlib.axes import Axes

from psdm_analysis.models.grid_with_results import GridWithResults
from psdm_analysis.models.input.connector.lines import Lines
from psdm_analysis.models.result.grid.connector import ConnectorResult
from psdm_analysis.models.result.grid.transformer import Transformer2WResult
from psdm_analysis.plots.common.line_plot import ax_plot_time_series
from psdm_analysis.plots.common.utils import FIGSIZE, ax_plot_secondary_axis, set_title


def plot_transformer_s_rated(
    res: Transformer2WResult,
    side: str,
    gwr: GridWithResults,
    resolution: str,
    include_utilisation: bool = True,
    show_util_grid_lines: bool = False,
    title: Optional[str] = None,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    transformer_name = res.name if res.name else res.input_model
    if title is None:
        title = f"Rated Power Transformer: {transformer_name}"
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_transformer_s_rated(
        ax,
        res,
        side,
        gwr,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    if include_utilisation:
        ax_plot_secondary_axis(
            ax,
            res.calc_transformer_utilisation(gwr),
            "Transformer Utilisation",
            show_secondary_grid_lines=show_util_grid_lines,
        )
    ax.set_ylabel("Rated Power in kVA")
    set_title(ax, title)
    return fig, ax


def ax_plot_transformer_s_rated(
    ax: Axes,
    res: Transformer2WResult,
    side: str,
    gwr: GridWithResults,
    resolution: str,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.i_a_mag) == 0:
        raise ValueError("Transformer current time series is empty. No data to plot")

    rated_power = res.calc_rated_power_gwr(gwr, side)
    ax_plot_time_series(
        ax,
        rated_power,
        res.type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Rated Power in kVA")


def plot_line_current(
    res: ConnectorResult,
    side: str,
    resolution: str,
    include_utilisation: bool = False,
    i_max_src: Optional[Union[float, Lines]] = None,
    title: Optional[str] = None,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    line_name = res.name if res.name else res.input_model
    if title is None:
        title = f"Current Magnitude Line: {line_name}"
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_line_current(
        ax,
        res,
        side,
        resolution,
        i_max_src=i_max_src,
        include_utilisation=include_utilisation,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Current in Ampere")
    set_title(ax, title)
    return fig, ax


def ax_plot_line_current(
    ax: Axes,
    res: ConnectorResult,
    side: str,
    resolution: str,
    include_utilisation: bool = True,
    i_max_src: Optional[Union[float, Lines]] = None,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.i_a_mag) == 0:
        raise ValueError("Line current time series is empty. No data to plot")

    current = get_connector_current(side, res)

    ax = ax_plot_time_series(
        ax,
        current,
        res.type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Current Magnitude in Ampere")

    if include_utilisation:
        if i_max_src is None:
            logging.warning("No i_max_src provided. Cannot plot utilisation")
            return
        line_i_max = (
            i_max_src
            if isinstance(i_max_src, float)
            else i_max_src.subset(res.input_model).i_max.iloc[0]
        )
        utilisation = current / line_i_max
        ax_plot_secondary_axis(
            ax, utilisation, "Line Utilisation", show_secondary_grid_lines=True
        )


def plot_connector_current(
    res: ConnectorResult,
    side: str,
    resolution: str,
    title: Optional[str] = None,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    connector_type = res.type.get_plot_name()
    connector_name = res.name if res.name else res.input_model
    if title is None:
        title = f"Current Magnitude {connector_type}: {connector_name}"
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_connector_current(
        ax,
        res,
        side,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Current in Ampere")
    set_title(ax, title)
    return fig, ax


def ax_plot_connector_current(
    ax: Axes,
    res: ConnectorResult,
    side: str,
    resolution: str,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    connector_type = res.type.get_plot_name()
    if len(res.i_a_mag) == 0:
        raise ValueError(
            f"{connector_type} current time series is empty. No data to plot"
        )

    current = get_connector_current(side, res)

    ax = ax_plot_time_series(
        ax,
        current,
        res.type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Current Magnitude in Ampere")


def get_connector_current(side: str, res: ConnectorResult):
    if side == "a":
        return res.i_a_mag
    elif side == "b":
        return res.i_b_mag
    else:
        raise ValueError('Side should be either "a" for node a or "b" for node b')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['psdm_analysis',
 'psdm_analysis.analysis',
 'psdm_analysis.conversion',
 'psdm_analysis.io',
 'psdm_analysis.models',
 'psdm_analysis.models.input',
 'psdm_analysis.models.input.connector',
 'psdm_analysis.models.input.container',
 'psdm_analysis.models.input.participant',
 'psdm_analysis.models.result',
 'psdm_analysis.models.result.grid',
 'psdm_analysis.models.result.participant',
 'psdm_analysis.plots',
 'psdm_analysis.plots.common',
 'psdm_analysis.plots.results',
 'psdm_analysis.processing']

package_data = \
{'': ['*']}

install_requires = \
['jupyter>=1.0.0,<2.0.0',
 'matplotlib>=3.6.0,<4.0.0',
 'networkx>=3.1,<4.0',
 'numpy>=1.24.1,<2.0.0',
 'openpyxl>=3.1.2,<4.0.0',
 'pandas==2.0.1',
 'plotly>=5.6.0,<6.0.0',
 'pyarrow>=11.0.0,<12.0.0',
 'python-dotenv>=1.0.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'scipy>=1.10.0,<2.0.0',
 'seaborn>=0.12.1,<0.13.0',
 'shapely>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'psdm-analysis',
    'version': '0.1.5',
    'description': '',
    'long_description': '# psdm-analysis\n\nThe psdm-analysis tool is meant to parse the [Power System Data Model (PSDM)](https://github.com/ie3-institute/PowerSystemDataModel) as well as provide calculation and plotting utilities to analyze the respective data.\n\nIt is currently under development and therefore highly unstable. So if you want to use it, expect it to change quite frequently for now.',
    'author': 'Thomas Oberliessen',
    'author_email': 'thomas.oberliessen@googlemail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>3.10,<3.12',
}


setup(**setup_kwargs)



from dataclasses import dataclass

import pandas as pd
from pandas import DataFrame, Series

from psdm_analysis.models.input.enums import RawGridElementsEnum
from psdm_analysis.models.result.grid.node import NodeResult, NodesResult
from psdm_analysis.models.result.power import PQResult


@dataclass(frozen=True)
class EnhancedNodeResult(NodeResult):
    def __eq__(self, other):
        if not isinstance(other, EnhancedNodeResult):
            return False
        return (
            (self.input_model == other.input_model)
            & (self.name == other.name)
            & (self.data.equals(other.data))
        )

    @classmethod
    def from_node_result(
        cls, node_res: NodeResult, pq: PQResult
    ) -> "EnhancedNodeResult":
        return cls(
            RawGridElementsEnum.NODE,
            node_res.input_model,
            node_res.name,
            pd.concat([node_res.data, pq.data], axis=1).sort_index(),
        )

    @property
    def p(self) -> Series:
        return self.data["p"]

    @property
    def q(self) -> Series:
        return self.data["q"]


@dataclass(frozen=True)
class EnhancedNodesResult(NodesResult):
    entities: dict[str, EnhancedNodeResult]

    @classmethod
    def from_nodes_result(
        cls, nodes_result: NodesResult, nodal_pq: dict[str, PQResult]
    ):
        return cls(
            RawGridElementsEnum.NODE,
            {
                uuid: EnhancedNodeResult.from_node_result(result, nodal_pq[uuid])
                for uuid, result in nodes_result.entities.items()
            },
        )

    @property
    def p(self) -> DataFrame:
        return (
            pd.concat(
                [
                    node_res.p.rename(node_res.input_model)
                    for node_res in self.entities.values()
                ],
                axis=1,
            )
            .sort_index()
            .ffill()
            .fillna(0)
        )

    @property
    def q(self) -> DataFrame:
        return (
            pd.concat(
                [
                    node_res.q.rename(node_res.input_model)
                    for node_res in self.entities.values()
                ],
                axis=1,
            )
            .sort_index()
            .ffill()
            .fillna(0)
        )



import json

import plotly.graph_objs as go
from pandas import Series
from shapely.geometry import LineString

from psdm_analysis.models.input.container.grid_container import GridContainer
from psdm_analysis.plots.common.utils import BLUE, GREY, RED, rgb_to_hex


def get_lons_lats(geojson: str):
    coordinates = json.loads(geojson)["coordinates"]
    return list(zip(*coordinates))  # returns lons, lats


def grid_plot(grid: GridContainer):
    """
    Plots the grid on an OpenStreetMap. Lines that are disconnected due to open switches will be grey.

    ATTENTION:
    We currently consider the node_b of the switches to be the auxiliary switch node.
    This is not enforced within the PSDM so might not work as expected.
    If that is the case the wrong lines might be grey.
    """
    fig = go.Figure()

    # Get disconnected lines via opened switches
    opened_switches = grid.raw_grid.switches.get_opened()

    disconnected_lines = grid.raw_grid.lines.find_lines_by_nodes(opened_switches.node_b)
    _, connected_lines = grid.raw_grid.lines.subset_split(disconnected_lines.uuids)

    connected_lines.data.apply(lambda line: add_line_trace(fig, line), axis=1)
    disconnected_lines.data.apply(
        lambda line: add_line_trace(fig, line, is_disconnected=True), axis=1
    )

    add_node_trace(fig, grid)

    center_lat = grid.raw_grid.nodes.data["latitude"].mean()
    center_lon = grid.raw_grid.nodes.data["longitude"].mean()

    # Dynamically calculate the zoom level
    lat_range = (
        grid.raw_grid.nodes.data["latitude"].max()
        - grid.raw_grid.nodes.data["latitude"].min()
    )
    lon_range = (
        grid.raw_grid.nodes.data["longitude"].max()
        - grid.raw_grid.nodes.data["longitude"].min()
    )

    zoom = 12 - max(lat_range, lon_range)

    fig.update_layout(
        # mapbox = {"zoom"=10},
        showlegend=False,
        mapbox_style="open-street-map",
        margin={"r": 0, "t": 0, "l": 0, "b": 0},
        mapbox=dict(
            center=dict(lat=center_lat, lon=center_lon),
            zoom=zoom,  # Adjust the zoom level as per the calculated heuristic
            style="open-street-map",
        ),
    )

    return fig


def add_line_trace(fig: go.Figure, line_data: Series, is_disconnected: bool = False):
    lons, lats = get_lons_lats(line_data.geo_position)
    hover_text = line_data["id"]

    color = GREY if is_disconnected else RED

    # Add the lines
    fig.add_trace(
        go.Scattermapbox(
            mode="lines",
            lon=lons,
            lat=lats,
            hoverinfo="skip",  # Skip hoverinfo for the lines
            line=dict(color=rgb_to_hex(color)),
        ),
    )

    # Create a LineString object from the line's coordinates
    line = LineString(zip(lons, lats))

    # Calculate the midpoint on the line based on distance
    midpoint = line.interpolate(line.length / 2)

    # Add a transparent marker at the midpoint of the line for hover text
    fig.add_trace(
        go.Scattermapbox(
            mode="markers",
            lon=[midpoint.x],
            lat=[midpoint.y],
            hoverinfo="text",
            hovertext=hover_text,
            marker=dict(size=0, opacity=0, color=rgb_to_hex(RED)),
        )
    )


def add_node_trace(fig: go.Figure, grid: GridContainer):
    node_hover_data = grid.get_nodal_sp_count_and_power()
    nodes_data = grid.raw_grid.nodes.data

    def to_hover_text(node_data: Series):
        return (
            node_data["id"]
            + "<br>"
            + "<br>".join(
                [
                    f"{key}={value}"
                    for key, value in node_hover_data[node_data.name].items()
                ]
            )
        )

    text = nodes_data.apply(
        lambda node_data: to_hover_text(node_data), axis=1
    ).to_list()

    fig.add_trace(
        go.Scattermapbox(
            mode="markers",
            lon=nodes_data["longitude"],
            lat=nodes_data["latitude"],
            hovertext=text,
            hoverinfo="text",
            marker=dict(size=6, color=rgb_to_hex(BLUE)),
            text=text,
        )
    )



from dataclasses import dataclass
from typing import List

from psdm_analysis.models.input.enums import SystemParticipantsEnum
from psdm_analysis.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class PhotovoltaicPowerPlants(SystemParticipants):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.PHOTOVOLTAIC_POWER_PLANT

    @property
    def s_rated(self):
        return self.data["s_rated"]

    @property
    def albedo(self):
        return self.data["albedo"]

    @property
    def azimuth(self):
        return self.data["azimuth"]

    @property
    def elevation_angle(self):
        return self.data["elevation_angle"]

    @property
    def k_g(self):
        return self.data["k_g"]

    @property
    def k_t(self):
        return self.data["k_t"]

    @property
    def market_reaction(self):
        return self.data["market_reaction"]

    @property
    def cos_phi_rated(self):
        return self.data["cos_phi_rated"]

    @staticmethod
    def attributes() -> List[str]:
        return SystemParticipants.attributes() + [
            "albedo",
            "azimuth",
            "elevation_angle",
            "k_g",
            "k_t",
            "market_reaction",
            "cos_phi_rated",
        ]



from datetime import datetime

import pandas as pd
from pandas import DataFrame


def divide_positive_negative(df: DataFrame):
    positive = df.copy()
    positive[positive < 0] = 0
    negative = df.copy()
    negative[negative > 0] = 0
    return positive, negative


def join_dataframes(dfs: list[DataFrame]):
    return pd.concat(dfs)


def filter_data_for_time_interval(
    data: DataFrame, start: datetime, end: datetime
) -> DataFrame:
    if data.empty:
        return data
    if not data.index.is_monotonic_increasing:
        data = data.sort_index()
    # we want the first value to be start if present and if not the last value before start
    all_after_start = (data.index > start).all()
    all_after_end = (data.index >= end).all()
    if all_after_start and all_after_end:
        return data.drop(data.index)
    elif all_after_start:
        start_row = data.index[1]
    else:
        before_start = data.index[data.index <= start]
        start_row = before_start[-1]
    # since this is the start of the time interval
    end_row = data.index[data.index <= end][-1]
    filtered_data = data[start_row:end_row]
    return filtered_data.rename(index={filtered_data.loc[start_row].name: start})



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import List

import numpy as np
import pandas as pd
from pandas import DataFrame, Series

from psdm_analysis.models.entity import ResultEntities
from psdm_analysis.models.result.grid.node import NodeResult
from psdm_analysis.models.result.participant.dict import ResultDict


@dataclass(frozen=True)
class ConnectorResult(ResultEntities):
    @staticmethod
    def attributes() -> List[str]:
        return ["i_a_ang", "i_a_mag", "i_b_ang", "i_b_mag"]

    @property
    def i_a_ang(self) -> Series:
        return self.data["i_a_ang"]

    @property
    def i_a_mag(self) -> Series:
        return self.data["i_a_mag"]

    @property
    def i_a_complex(self) -> Series:
        return self.i_a_mag * np.exp(1j * np.radians(self.i_a_ang))

    @property
    def i_b_ang(self) -> Series:
        return self.data["i_b_ang"]

    @property
    def i_b_mag(self) -> Series:
        return self.data["i_b_mag"]

    @property
    def i_b_complex(self) -> Series:
        return self.i_b_mag * np.exp(1j * np.radians(self.i_b_ang))

    # todo: Check if calculation is correct
    def calc_rated_power(
        self, node_res: NodeResult, voltage_level_kv: float, node="a", absolute=True
    ):  # in kVA
        s_rated_fn = (
            lambda i_complex: math.sqrt(3)
            * node_res.v_complex(voltage_level_kv)
            * np.conj(i_complex)
        )
        if node == "a":
            res = s_rated_fn(self.i_a_complex)
        elif node == "b":
            res = s_rated_fn(self.i_b_complex)
        else:
            raise ValueError("Node must be either hv or lv")
        return np.absolute(res) if absolute else res


@dataclass(frozen=True)
class ConnectorsResult(ResultDict):
    entities: dict[str, ConnectorResult]

    @property
    def i_a_ang(self) -> DataFrame:
        return pd.concat(
            [
                node_res.i_a_ang.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )

    @property
    def i_a_mag(self) -> DataFrame:
        return pd.concat(
            [
                node_res.i_a_mag.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )

    @property
    def i_b_ang(self) -> DataFrame:
        return pd.concat(
            [
                node_res.i_b_ang.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )

    @property
    def i_b_mag(self) -> DataFrame:
        return pd.concat(
            [
                node_res.i_b_mag.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )
